import cvxpy as cp
import numpy as np
import pytest

from dsp import conjugate, inner, saddle_max, saddle_min, weighted_log_sum_exp
from dsp.local import LocalVariable, LocalVariableError
from dsp.parser import DSPError
from dsp.problem import MinimizeMaximize, SaddlePointProblem, is_dsp
from dsp.semi_infinite_canon import saddle_max_canon, saddle_min_canon


def test_semi_infinite_matrix():
    x = cp.Variable(2, name="x", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)

    x_d = LocalVariable(2, name="x_d", nonneg=True)
    y_d = LocalVariable(2, name="y_d", nonneg=True)

    A = np.array([[1, 2], [3, 1]])
    inner_expr = inner(x, A @ y)

    # Saddle problem
    saddle_obj = MinimizeMaximize(inner_expr)
    saddle_problem = SaddlePointProblem(saddle_obj, [cp.sum(x) == 1, cp.sum(y) == 1])
    saddle_problem.solve()

    assert np.isclose(saddle_problem.value, 1.66666, atol=1e-4)
    assert np.allclose(x.value, [0.66666, 0.33333], atol=1e-4)
    assert np.allclose(y.value, [0.33333, 0.66666], atol=1e-4)

    # Saddle max problem

    inner_expr = inner(x, A @ y_d)

    obj = cp.Minimize(saddle_max(inner_expr, [cp.sum(y_d) == 1]))
    constraints = [cp.sum(x) == 1]

    problem = cp.Problem(obj, constraints)
    problem.solve(solver=cp.SCS)

    assert np.isclose(problem.value, 1.66666, atol=1e-4)

    # Saddle min problem

    inner_expr = inner(x_d, A @ y)

    f = saddle_min(inner_expr, [cp.sum(x_d) == 1])
    obj = cp.Maximize(f)
    constraints = [cp.sum(y) == 1]

    problem = cp.Problem(obj, constraints)
    problem.solve(solver=cp.SCS)
    assert np.isclose(saddle_problem.value, 1.66666, atol=1e-4)
    assert np.allclose(x.value, [0.66666, 0.33333], atol=1e-4)
    assert np.allclose(y.value, [0.33333, 0.66666], atol=1e-4)


def test_dcp_concave_max_and_dummy():
    x = cp.Variable(2, name="x", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)

    A = np.array([[1, 2], [3, 4]])
    inner_expr = inner(x, A @ y)

    with pytest.raises(LocalVariableError):
        saddle_max(inner_expr, [cp.sum(y) == 1])

    y2 = LocalVariable(2, name="y", nonneg=True)

    A = np.array([[1, 2], [3, 4]])
    inner_expr = inner(x, A @ y2)

    f_max = saddle_max(inner_expr, [cp.sum(y2) == 1])

    obj = cp.Maximize(f_max)
    assert not obj.is_dcp()


def test_semi_infinite_expr():
    x = cp.Variable(2, name="x", nonneg=True)
    y_dummy = LocalVariable(2, name="y_dummy", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)

    wlse = weighted_log_sum_exp(x, y)

    # Trying to create a saddle_max with a variable for y (instead of a dummy)
    with pytest.raises(LocalVariableError):
        saddle_max(2 * wlse + y[1] + cp.exp(x[1]), [y <= 1])

    wlse = weighted_log_sum_exp(x, y_dummy)

    # creating a valid saddle_max with a dummy variable
    sup_y_f = saddle_max(2 * wlse + y_dummy[1] + cp.exp(x[1]), [y_dummy <= 1])
    assert sup_y_f.is_dcp()
    assert sup_y_f.is_dsp()

    # trying to use the same dummy variable in a new SE raises an error
    with pytest.raises(LocalVariableError):
        saddle_max(2 * wlse + 2 * cp.sum(y_dummy), [y_dummy <= 1])

    # trying to get the value of the saddle_max before x has a value returns None
    assert sup_y_f.numeric(values=np.ones(1)) is None

    # trying to get the value of y_dummy before x has a value returns None
    assert y_dummy.value is None

    x.value = np.ones(2)

    assert np.allclose(y_dummy.value, np.ones(2))

    val = sup_y_f.numeric(values=np.ones(1))
    assert np.isclose(val, 2 * np.log(2 * np.e) + 1 + np.e)


def test_multiple_dummies():
    x = cp.Variable(2, name="x", nonneg=True)
    y1_local = LocalVariable(name="y1_local", nonneg=True)
    y2_local = LocalVariable(name="y2_local", nonneg=True)

    y = cp.Variable(name="y", nonneg=True)

    wlse = weighted_log_sum_exp(x, cp.hstack([y1_local, y2_local]))

    # only one dummy variable is specified. This raises a DSP error since not
    # all concave variables are specified, and not a local variable error,
    # because all of the concave variables are specified as local variables.
    sup_y_f = saddle_max(2 * wlse + y1_local + cp.exp(x[1]), [y1_local <= 1])
    assert sup_y_f.is_dcp()
    assert sup_y_f.is_dsp()

    # trying a mix of dummy and variable. This raises a local variable error on
    # construction.
    with pytest.raises(LocalVariableError):
        sup_y_f = saddle_max(2 * wlse + y + cp.exp(x[1]), [y1_local <= 1, y <= 1])

    y1_local = LocalVariable(name="y1_local", nonneg=True)
    y2_local = LocalVariable(name="y2_local", nonneg=True)
    wlse = weighted_log_sum_exp(x, cp.hstack([y1_local, y2_local]))
    sup_y_f = saddle_max(2 * wlse + y1_local + cp.exp(x[1]), [y1_local <= 1, y2_local <= 1])

    assert sup_y_f.numeric(values=np.ones(1)) is None
    assert y1_local.value is None
    assert y2_local.value is None

    x.value = np.ones(2)

    assert np.allclose(y1_local.value, np.ones(1))
    assert np.allclose(y2_local.value, np.ones(1))


def test_trivial_se():
    x = cp.Variable(name="x", nonneg=True)

    f = cp.exp(x)

    # trivial saddle max with no dummy variable
    F = saddle_max(f, [])

    prob = cp.Problem(cp.Minimize(F), [x >= 1])
    prob.solve()

    assert np.isclose(prob.value, np.e)

    y = cp.Variable(name="y", nonneg=True)
    f = cp.log(y)

    # trivial saddle min with no dummy variable
    F = saddle_min(f, [])

    prob = cp.Problem(cp.Maximize(F), [y <= 1])
    prob.solve()

    assert np.isclose(prob.value, 0)


def test_nested_saddle():
    x = cp.Variable(2, name="x", nonneg=True)
    y_1 = LocalVariable(2, name="y_1", nonneg=True)
    y_2 = LocalVariable(name="y_2", nonneg=True)

    f = weighted_log_sum_exp(x, y_1)

    sup_y_f = saddle_max(f, [y_1 <= 1])

    g = weighted_log_sum_exp(sup_y_f, y_1[1])
    with pytest.raises(LocalVariableError):  # adding a local variable to multiple SEs
        saddle_max(g, [y_1 <= 1])

    g = weighted_log_sum_exp(sup_y_f, y_2)
    sup_y_g = saddle_max(g, [y_2 <= 1])

    prob = cp.Problem(cp.Minimize(sup_y_g), [x >= 1])
    prob.solve()

    assert np.isclose(prob.value, np.log(2 * np.e))

    # TODO: prevent local variables from appearing in dcp/dsp problems


def test_saddle_max():
    x1_local = LocalVariable(name="x1", nonneg=True)
    x2_local = LocalVariable(name="x2", nonneg=True)
    y = cp.Variable(2, name="y", nonneg=True)

    x = cp.Variable(name="x", nonneg=True)

    wlse = weighted_log_sum_exp(cp.hstack([x1_local, x2_local]), y)

    # only one dummy variable is specified
    inf_x_f = saddle_min(2 * wlse + x1_local + cp.log(y[1]), [x1_local >= 1])
    assert inf_x_f.is_dcp()
    assert inf_x_f.is_dsp()

    # trying a mix of dummy and variable
    with pytest.raises(LocalVariableError):
        saddle_min(2 * wlse + x + cp.log(y[1]), [x1_local >= 1, x >= 1])

    x1_local = LocalVariable(name="x1", nonneg=True)
    x2_local = LocalVariable(name="x2", nonneg=True)
    wlse = weighted_log_sum_exp(cp.hstack([x1_local, x2_local]), y)
    inf_x_f = saddle_min(2 * wlse + x1_local + cp.log(y[1]), [x1_local >= 1, x2_local >= 1])

    assert inf_x_f.numeric(values=np.ones(2)) is None
    assert x1_local.value is None
    assert x2_local.value is None

    y.value = np.ones(2)

    assert np.isclose(x1_local.value, 1)
    assert np.isclose(x2_local.value, 1)


def test_non_dsp_saddle():
    x = cp.Variable(2, name="x", nonneg=True)
    y = LocalVariable(2, name="y", nonneg=True)

    f = weighted_log_sum_exp(x, y) + cp.exp(y[1])

    sup_y_f = saddle_max(f, [y <= 1])

    assert not sup_y_f.is_dsp()

    prob = cp.Problem(cp.Minimize(sup_y_f), [x >= 1])

    assert not is_dsp(prob)
    assert not prob.is_dcp()

    # currently a DCPError is raised. Do we want a DCPError or a DSPError?
    with pytest.raises(cp.error.DCPError):
        prob.solve()


def test_dsp_canon_error():
    x = cp.Variable(2, name="x", nonneg=True)
    y = LocalVariable(2, name="y", nonneg=True)

    f = weighted_log_sum_exp(x, y) + cp.exp(y[1])

    sup_y_f = saddle_max(f, [y <= 1])

    prob = cp.Problem(cp.Minimize(sup_y_f), [x >= 1])

    with pytest.raises(cp.error.DCPError):
        prob.solve()


def test_saddle_no_constraint():
    x = cp.Variable(2, name="x", nonneg=True)
    y = LocalVariable(2, name="y", nonneg=True)

    # f = weighted_log_sum_exp(x, y)
    f = inner(x, y)

    sup_y_f = saddle_max(f, [])

    assert sup_y_f.is_dcp()
    assert sup_y_f.is_dsp()

    prob = cp.Problem(cp.Minimize(sup_y_f), [x >= -1e2])
    assert prob.is_dcp()
    prob.solve()


def test_conj():
    x = LocalVariable(name="x_local")
    f = 0.5 * cp.square(x)

    f_star = conjugate(f)

    assert f_star.is_dcp()
    assert f_star.name() == "conjugate(0.5 @ power(x_local, 2.0))"

    f_star.variables()[0].value = 2

    obj = cp.Minimize(f_star)
    prob = cp.Problem(obj, [f_star.variables()[0] >= 1])

    prob.solve()
    assert np.isclose(prob.value, 0.5)


def test_canon_non_dsp():
    x = cp.Variable(name="x")
    y = LocalVariable(name="y")
    f = cp.exp(x) * cp.log(y)

    with pytest.raises(DSPError):
        saddle_max_canon(f, None)

    with pytest.raises(DSPError):
        saddle_min_canon(f, None)



import cvxpy as cp
import numpy as np
import pytest

from dsp import MinimizeMaximize, inner, saddle_inner, saddle_max
from dsp.local import LocalVariable
from dsp.problem import SaddlePointProblem


def test_ubounded_domains_exp():
    x = cp.Variable(2, name="x")
    y = cp.Variable(2, name="y", nonneg=True)

    with pytest.warns(UserWarning, match="Gy is non-positive"):
        f = -saddle_inner(cp.exp(x), cp.log(y))

    saddle_problem = SaddlePointProblem(
        # MinimizeMaximize(f), [cp.sum(y) >= 4, x <= 1]
        MinimizeMaximize(f),
        [],  # -> fatal mosek error if y not nonneg
    )  # , x >= 1, cp.sum(x) == 2])

    saddle_problem.solve()
    assert np.isclose(saddle_problem.value, 0.0, atol=1e-6)


def test_unbounded_domains_inv_pos():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")

    # f_cvx = cp.inv_pos(x)
    # f_ccv = -cp.inv_pos(y)+2

    a = 1.2
    f_cvx = cp.power(x, -a)
    f_ccv = -cp.power(y, -a) + 2

    with pytest.warns(UserWarning, match="Gy is non-positive"):
        f = saddle_inner(f_cvx, f_ccv)

    saddle_problem = SaddlePointProblem(MinimizeMaximize(f), [x >= 1, y >= 1])

    # saddle_problem.solve(solver=cp.MOSEK)
    # x.value, y.value

    y_local = LocalVariable(name="y")
    f_cvx = cp.inv_pos(x) + 0.00001
    f_ccv = cp.power(y_local, 0.01)
    f = saddle_inner(f_cvx, f_ccv)

    F = saddle_max(f, [y_local >= 1])
    prob = cp.Problem(cp.Minimize(F), [x >= 1])
    prob.solve()
    prob.value
    x.value, y.value

    # for x_val in [1, 1.1,1.3,1.5,2]:
    #     x.value = x_val
    #     print(f"{x_val=}, {F.value=}")

    # print()


def test_lagrangian():
    x = cp.Variable(name="x")
    nu = cp.Variable(name="y")

    f = cp.exp(x) + inner(x - 1, nu)

    saddle_problem = SaddlePointProblem(MinimizeMaximize(f), [x >= -10, 0 * nu == 0])
    saddle_problem.solve()
    assert np.isclose(saddle_problem.value, np.e)



from __future__ import annotations

import itertools
from typing import Iterable

import cvxpy as cp
from cvxpy import multiply
from cvxpy.atoms.affine.add_expr import AddExpression
from cvxpy.atoms.affine.binary_operators import DivExpression, MulExpression
from cvxpy.atoms.affine.unary_operators import NegExpression
from cvxpy.constraints.constraint import Constraint

import dsp
from dsp.cone_transforms import (
    K_repr_ax,
    K_repr_by,
    KRepresentation,
    LocalToGlob,
    split_K_repr_affine,
)


class DSPError(Exception):
    pass


class AffineDSPError(DSPError):
    def __init__(self, message: str, affine_vars: set[cp.Variable]) -> None:
        super().__init__(message)
        self.affine_vars = affine_vars


def affine_error_message(affine_vars: list[cp.Variable]) -> str:
    return (
        f"Cannot resolve curvature of variables {[v.id for v in affine_vars]}. "
        f"Specify curvature of these variables as "
        f"SaddleProblem(obj, constraints, minimization_vars, maximization_vars)."
    )


class Parser:
    def __init__(self, convex_vars: set[cp.Variable], concave_vars: set[cp.Variable]) -> None:

        self.convex_vars: set[cp.Variable] = convex_vars
        self.concave_vars: set[cp.Variable] = concave_vars
        if self.convex_vars & self.concave_vars:
            raise DSPError("Cannot have variables in both maximization_vars and minimization_vars.")

        self.affine_vars: set[cp.Variable] = set()
        self._x_constraints = None
        self._y_constraints = None

    @property
    def x_constraints(self) -> list[Constraint]:
        if self._x_constraints is None:
            return []
        return self._x_constraints

    @property
    def y_constraints(self) -> list[Constraint]:
        if self._y_constraints is None:
            return []
        return self._y_constraints

    def add_to_convex_vars(self, variables: Iterable[cp.Variable]) -> None:
        variables = set(variables)
        if variables & self.concave_vars:
            raise DSPError("Cannot add variables to both convex and concave set.")
        self.affine_vars -= variables
        self.convex_vars |= variables

    def add_to_concave_vars(self, variables: Iterable[cp.Variable]) -> None:
        variables = set(variables)
        if variables & self.convex_vars:
            raise DSPError("Cannot add variables to both convex and concave set.")

        self.affine_vars -= variables
        self.concave_vars |= variables

    def parse_scalar_mul(
        self, expr: cp.Expression, switched: bool, repr_parse: bool, **kwargs: dict
    ) -> KRepresentation | None:
        assert expr.args[0].is_constant() or expr.args[1].is_constant()
        const_ind = 0 if isinstance(expr.args[0], cp.Constant) else 1
        var_ind = 1 - const_ind

        s = expr.args[const_ind]
        var_expr = expr.args[var_ind]

        if s.is_nonneg():
            return_val = self._parse_expr(var_expr, switched, repr_parse, **kwargs)
        else:
            return_val = self._parse_expr(var_expr, not switched, repr_parse, **kwargs)

        if repr_parse:
            assert return_val is not None
            return return_val.scalar_multiply(abs(s.value))

    def parse_div(
        self, expr: DivExpression, switched: bool, repr_parse: bool, **kwargs: dict
    ) -> KRepresentation | None:
        assert expr.args[1].is_constant()
        divisor = expr.args[1]
        return self.parse_scalar_mul(expr.args[0] * (1 / divisor), switched, repr_parse, **kwargs)

    def parse_known_curvature_repr(
        self, expr: cp.Expression, local_to_glob: LocalToGlob
    ) -> KRepresentation:
        if set(expr.variables()) <= self.convex_vars:
            assert expr.is_convex()
            return K_repr_ax(expr)
        elif set(expr.variables()) <= self.concave_vars:
            assert expr.is_concave()
            return K_repr_by(expr, local_to_glob)
        else:
            raise ValueError

    def parse_known_curvature_vars(self, expr: cp.Expression, switched: bool) -> None:
        variables = expr.variables()
        if expr.is_affine():
            self.affine_vars |= set(expr.variables()) - self.convex_vars - self.concave_vars
        elif expr.is_convex():
            self.add_to_convex_vars(variables) if not switched else self.add_to_concave_vars(
                variables
            )
        elif expr.is_concave():
            self.add_to_concave_vars(variables) if not switched else self.add_to_convex_vars(
                variables
            )
        else:
            raise ValueError(f"Cannot parse {expr=} with {expr.curvature=}.")

    def parse_add(
        self, expr: cp.Expression, switched: bool, repr_parse: bool, **kwargs: dict
    ) -> KRepresentation | None:
        assert isinstance(expr, AddExpression)
        if repr_parse:
            K_reprs = [self.parse_expr_repr(arg, switched, **kwargs) for arg in expr.args]
            return KRepresentation.sum_of_K_reprs(K_reprs)
        else:
            for arg in expr.args:
                self.parse_expr_variables(arg, switched, **kwargs)

    def parse_dsp_atom(
        self, expr: dsp.saddle_atoms.SaddleAtom, switched: bool, repr_parse: bool, **kwargs: dict
    ) -> KRepresentation | None:
        assert isinstance(expr, dsp.saddle_atoms.SaddleAtom)
        if repr_parse:
            return expr.get_K_repr(**kwargs, switched=switched)
        else:
            convex_vars = expr.convex_variables() if not switched else expr.concave_variables()
            concave_vars = expr.concave_variables() if not switched else expr.convex_variables()
            self.add_to_convex_vars(convex_vars)
            self.add_to_concave_vars(concave_vars)

    def parse_bilin(
        self, expr: cp.Expression, switched: bool, repr_parse: bool, **kwargs: dict
    ) -> KRepresentation | None:
        if repr_parse:
            raise ValueError("Unexpected; should fail in variable parse.")
        else:
            if all(arg.is_affine() for arg in expr.args):
                raise DSPError("Use inner instead for bilinear forms.")
            else:
                raise DSPError("Use saddle_inner instead.")

    def parse_expr_variables(self, expr: cp.Expression, switched: bool, **kwargs: dict) -> None:
        self._parse_expr(expr, switched, repr_parse=False, **kwargs)

    def parse_expr_repr(
        self, expr: cp.Expression, switched: bool, local_to_glob: LocalToGlob
    ) -> KRepresentation:

        K_repr = self._parse_expr(expr, switched, repr_parse=True, local_to_glob=local_to_glob)
        assert isinstance(K_repr, KRepresentation)
        return K_repr

    @staticmethod
    def contains_curvature_lumping(expr: cp.Expression) -> bool:
        """
        Recursively check if expression contains curvature lumping, i.e. assigning
        a non-affine curvature to an expression that contains a sum where some terms are affine.
        E.g., exp(x) + y is a convex expression, but we want to parse the affine part as affine.
        """
        if isinstance(expr, AddExpression):
            affine_term = any(expr_arg.is_affine() for expr_arg in expr.args)
            non_affine_term = any(not expr_arg.is_affine() for expr_arg in expr.args)
            return affine_term and non_affine_term
        else:
            return any(Parser.contains_curvature_lumping(expr_arg) for expr_arg in expr.args)

    def _parse_expr(
        self, expr: cp.Expression, switched: bool, repr_parse: bool, **kwargs: dict
    ) -> KRepresentation | None:

        # constant
        if not expr.variables():
            assert expr.size == 1
            return (
                KRepresentation.constant_repr(expr.value * (1 if not switched else -1))
                if repr_parse
                else None
            )

        # known curvature
        elif repr_parse and (
            (set(expr.variables()) <= self.convex_vars)
            or (set(expr.variables()) <= self.concave_vars)
        ):
            return self.parse_known_curvature_repr(expr * (1 if not switched else -1), **kwargs)

        elif (not repr_parse) and expr.is_dcp() and not self.contains_curvature_lumping(expr):
            return self.parse_known_curvature_vars(expr, switched)

        # convex and concave variables
        elif isinstance(expr, AddExpression):
            return self.parse_add(expr, switched, repr_parse, **kwargs)
        elif isinstance(expr, NegExpression):
            return self._parse_expr(expr.args[0], not switched, repr_parse, **kwargs)
        elif isinstance(expr, multiply):
            if expr.args[0].is_constant() or expr.args[1].is_constant():
                return self.parse_scalar_mul(expr, switched, repr_parse, **kwargs)
            else:
                return self.parse_bilin(expr, switched, repr_parse, **kwargs)
        elif isinstance(expr, dsp.saddle_atoms.SaddleAtom):
            return self.parse_dsp_atom(expr, switched, repr_parse, **kwargs)
        elif isinstance(expr, MulExpression):
            if expr.is_affine() and repr_parse:
                split_up_affine = split_K_repr_affine(expr, self.convex_vars, self.concave_vars)
                K_reprs = [self.parse_expr_repr(arg, switched, **kwargs) for arg in split_up_affine]
                return KRepresentation.sum_of_K_reprs(K_reprs)
            else:
                return self.parse_bilin(expr, switched, repr_parse, **kwargs)
        elif isinstance(expr, DivExpression):
            return self.parse_div(expr, switched, repr_parse, **kwargs)
        elif isinstance(expr, dsp.saddle_extremum.SaddleExtremum) and not expr.is_dsp():
            raise DSPError("Saddle min or max using non-DSP argument.")
        else:
            raise DSPError(f"Expression {expr} is not DSP.")


def _split_constraints(
    constraints: list[Constraint], parser: Parser
) -> tuple[list[Constraint], list[Constraint]]:
    n_constraints = len(constraints)
    x_constraints = []
    y_constraints = []

    while constraints:
        con_len = len(constraints)
        for c in list(constraints):
            c_vars = set(c.variables())
            if c_vars & parser.convex_vars:
                assert not (c_vars & parser.concave_vars)
                x_constraints.append(c)
                constraints.remove(c)
                parser.convex_vars |= c_vars
                parser.affine_vars -= c_vars
            elif c_vars & parser.concave_vars:
                assert not (c_vars & parser.convex_vars)
                y_constraints.append(c)
                constraints.remove(c)
                parser.concave_vars |= c_vars
                parser.affine_vars -= c_vars

        if con_len == len(constraints):
            raise DSPError(
                "Cannot split constraints, specify minimization_vars and maximization_vars"
            )

    assert len(x_constraints) + len(y_constraints) == n_constraints
    assert not (parser.convex_vars & parser.concave_vars)

    return x_constraints, y_constraints


def initialize_parser(
    expr: cp.Expression,
    minimization_vars: Iterable[cp.Variable],
    maximization_vars: Iterable[cp.Variable],
    constraints: list[Constraint] | None = None,
) -> Parser:
    parser = Parser(set(minimization_vars), set(maximization_vars))
    parser.parse_expr_variables(expr, switched=False)

    constraints = list(constraints)  # make copy

    x_constraints, y_constraints = _split_constraints(constraints, parser)

    parser._x_constraints = x_constraints
    parser._y_constraints = y_constraints

    if parser.affine_vars:
        raise AffineDSPError(affine_error_message(parser.affine_vars), parser.affine_vars)

    x_constraint_vars = set(
        itertools.chain.from_iterable(constraint.variables() for constraint in x_constraints)
    )
    y_constraint_vars = set(
        itertools.chain.from_iterable(constraint.variables() for constraint in y_constraints)
    )
    prob_vars = x_constraint_vars | y_constraint_vars | set(expr.variables())

    if prob_vars != (parser.convex_vars | parser.concave_vars):
        raise DSPError("Likely passed unused variables")
    return parser



from dsp.cvxpy_integration import add_is_dsp, extend_cone_canon_methods
from dsp.local import LocalVariable
from dsp.problem import MinimizeMaximize, SaddlePointProblem, is_dsp
from dsp.saddle_atoms import (
    inner,
    saddle_inner,
    saddle_quad_form,
    weighted_log_sum_exp,
    weighted_norm2,
)
from dsp.saddle_extremum import conjugate, saddle_max, saddle_min

__all__ = [
    "saddle_min",
    "saddle_max",
    "saddle_inner",
    "saddle_quad_form",
    "inner",
    "weighted_log_sum_exp",
    "MinimizeMaximize",
    "SaddlePointProblem",
    "is_dsp",
    "LocalVariable",
    "conjugate",
    "weighted_norm2",
]

extend_cone_canon_methods()
add_is_dsp()



from cvxpy import Expression, Problem

try:
    from cvxpy.reductions.dcp2cone.canonicalizers import CANON_METHODS
except ModuleNotFoundError:
    from cvxpy.reductions.dcp2cone.atom_canonicalizers import CANON_METHODS

from dsp.problem import is_dsp
from dsp.saddle_extremum import conjugate, saddle_max, saddle_min
from dsp.semi_infinite_canon import saddle_max_canon, saddle_min_canon


def extend_cone_canon_methods() -> None:
    """Extend the cone_canon_methods dictionary with methods from the
    semi_infinite_canon module.
    """

    CANON_METHODS.update(
        {
            saddle_max: saddle_max_canon,
            saddle_min: saddle_min_canon,
            conjugate: saddle_max_canon,
        }
    )


def add_is_dsp() -> None:
    """Add the is_dsp method to the cvxpy Problem class."""
    Problem.is_dsp = lambda self: is_dsp(self)
    Expression.is_dsp = lambda self: is_dsp(self)



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod

import cvxpy as cp
import numpy as np
from cvxpy.atoms.atom import Atom
from cvxpy.constraints import ExpCone
from cvxpy.utilities.sign import mul_sign

from dsp.cone_transforms import (
    K_repr_ax,
    K_repr_bilin,
    K_repr_by,
    K_repr_FxGy,
    KRepresentation,
    LocalToGlob,
    affine_to_canon,
    switch_convex_concave,
)
from dsp.parser import DSPError
from dsp.utils import np_vec


class SaddleAtom(Atom, ABC):
    def get_K_repr(self, local_to_glob: LocalToGlob, switched: bool = False) -> KRepresentation:
        if not self.is_dsp():
            raise DSPError(str(self) + " is not a DSP expression.")
            # TODO: descriptive DSP error messages
        return self._get_K_repr(local_to_glob, switched)

    @abstractmethod
    def _get_K_repr(self, local_to_glob: LocalToGlob, switched: bool) -> KRepresentation:
        raise NotImplementedError

    @abstractmethod
    def convex_variables(self) -> list[cp.Variable]:
        raise NotImplementedError

    @abstractmethod
    def concave_variables(self) -> list[cp.Variable]:
        raise NotImplementedError

    def affine_variables(self) -> list[cp.Variable]:
        known_curvature_vars = set(self.convex_variables() + self.concave_variables())
        return [v for v in self.variables() if v not in known_curvature_vars]

    @abstractmethod
    def get_convex_expression(self) -> cp.Expression:
        raise NotImplementedError

    @abstractmethod
    def get_concave_expression(self) -> cp.Expression:
        raise NotImplementedError

    @abstractmethod
    def is_dsp(self) -> bool:
        raise NotImplementedError

    def numeric(self, values: list[np.ndarray | None]) -> np.ndarray | None:
        if any([v is None for v in values]):
            return None
        return self._numeric(values)

    @abstractmethod
    def _numeric(self, values: list[np.ndarray]) -> np.ndarray:
        raise NotImplementedError

    def is_atom_convex(self) -> bool:
        return False

    def is_atom_concave(self) -> bool:
        return False

    def graph_implementation(self, arg_objs: list, shape: tuple[int, ...], data=None):  # noqa
        raise NotImplementedError

    def _grad(self, values: list):  # noqa
        raise NotImplementedError

    def shape_from_args(self) -> tuple[int, ...]:
        return ()

    def sign_from_args(self) -> tuple[bool, bool]:
        return (False, False)

    def is_incr(self, idx: int) -> bool:
        return False

    def is_decr(self, idx: int) -> bool:
        return False


class saddle_inner(SaddleAtom):
    def __init__(self, Fx: cp.Expression, Gy: cp.Expression) -> None:
        assert isinstance(Fx, cp.Expression)
        assert isinstance(Gy, cp.Expression)

        self.bilinear = Fx.is_affine() and Gy.is_affine()
        if (not self.bilinear) and (not Gy.is_nonneg()):
            warnings.warn(
                "Gy is non-positive. The y domain of saddle_inner is Gy >="
                " 0. The implicit constraint Gy >= 0 will be added to the problem."
            )

        self.Fx = Fx
        self.Gy = Gy

        assert len(Fx.shape) <= 1 or (
            len(Fx.shape) == 2 and min(Fx.shape) == 1
        )  # TODO: implement matrix inputs
        assert Fx.shape == Gy.shape or Fx.size == Gy.size

        super().__init__(Fx, Gy)

    def is_dsp(self) -> bool:
        x_cvx = self.Fx.is_convex()  # "Fx must be convex"
        x_nonneg = self.Fx.is_nonneg() or self.bilinear  # "Fx must be non-negative"
        y_ccv = self.Gy.is_concave()  # "Gy must be concave"
        return all([x_cvx, x_nonneg, y_ccv]) or self.bilinear

    def get_concave_expression(self) -> cp.Expression:
        Fx = self.Fx
        assert Fx.value is not None
        Fx = np_vec(Fx.value, order="F")
        Gy = cp.vec(self.Gy, order="F")
        return Fx @ Gy

    def get_convex_expression(self) -> cp.Expression:
        Gy = self.Gy
        assert Gy.value is not None
        Gy = np_vec(Gy.value, order="F")
        Fx = cp.vec(self.Fx, order="F")
        return Fx @ Gy

    def _get_K_repr(self, local_to_glob: LocalToGlob, switched: bool = False) -> KRepresentation:
        if self.bilinear:
            K_out = (
                K_repr_bilin(self.Fx, self.Gy, local_to_glob)
                if not switched
                else K_repr_bilin(-self.Gy, self.Fx, local_to_glob)
            )
        else:
            K_out = K_repr_FxGy(self.Fx, self.Gy, local_to_glob, switched)
            if not self.Gy.is_nonneg():
                outer_constraints = K_out.y_constraints if not switched else K_out.constraints
                outer_constraints.append(self.Gy >= 0)

        if not switched:
            K_out.concave_expr = lambda x: self.get_concave_expression()
        else:
            K_out.concave_expr = lambda x: -self.get_convex_expression()
        return K_out

    def _numeric(self, values: list[np.ndarray]) -> np.ndarray:
        return np_vec(values[0]) @ np_vec(values[1])

    def convex_variables(self) -> list[cp.Variable]:
        return self.Fx.variables()

    def concave_variables(self) -> list[cp.Variable]:
        return self.Gy.variables()

    def sign_from_args(self) -> tuple[bool, bool]:
        return (True, False)

    def is_incr(self, idx: int) -> bool:
        return True  # increasing in both arguments since y nonneg


class inner(saddle_inner):
    def __init__(self, Fx: cp.Expression, Gy: cp.Expression) -> None:
        assert isinstance(Fx, cp.Expression)
        assert isinstance(Gy, cp.Expression)

        assert (
            Fx.is_affine() and Gy.is_affine()
        ), "inner must be affine. Use convex-concave-inner for non-affine bilinear terms."

        super().__init__(Fx, Gy)

    def is_dsp(self) -> bool:
        x_aff = self.Fx.is_affine()
        y_aff = self.Gy.is_affine()
        return x_aff and y_aff

    def get_concave_expression(self) -> cp.Expression:
        return self.Fx.value @ self.Gy

    def get_convex_expression(self) -> cp.Expression:
        return self.Fx @ self.Gy.value

    def sign_from_args(self) -> tuple[bool, bool]:
        return mul_sign(self.args[0], self.args[1])

    def is_incr(self, idx: int) -> bool:
        return self.args[1 - idx].is_nonneg()

    def is_decr(self, idx: int) -> bool:
        return self.args[1 - idx].is_nonpos()


class weighted_log_sum_exp(SaddleAtom):
    def __init__(self, exponents: cp.Expression, weights: cp.Expression) -> None:
        """
        Implements the function f(x,y) = log(sum(exp(x_i) * y_i)) for vectors x and y.
        The weights, y, must be non-negative. If they are non recognized by
        cvxpy as nonnegative, an implicit domain constraint is added and a
        warning is provided.
        The exponents can be any convex cvxpy expression.

        The conic representation is:

            log sum_i y_i exp(x_i)
            min. f, t, u    f @ y + t
            subject to      f >= exp(x+u)
                            t >= -u-1

        """

        assert isinstance(exponents, cp.Expression)
        assert isinstance(weights, cp.Expression)

        if not weights.is_nonneg():
            warnings.warn(
                "Weights are non-positive. The domain of weighted log-sum-exp is y >="
                " 0. The implicit constraint y >= 0 will be added to"
                " the problem."
            )

        self.concave_composition = not weights.is_affine()

        self.exponents = exponents
        self.weights = weights

        assert len(exponents.shape) <= 1 or (
            len(exponents.shape) == 2 and min(exponents.shape) == 1
        )  # TODO: implement matrix inputs
        assert exponents.shape == weights.shape or exponents.size == weights.size

        super().__init__(exponents, weights)

    def _numeric(self, values: list[np.ndarray]) -> np.ndarray:
        x, y = values
        return np.log(np.sum(y * np.exp(x)))

    def is_dsp(self) -> bool:
        x_cvx = self.exponents.is_convex()
        y_ccv = self.weights.is_concave()
        return x_cvx and y_ccv

    def get_concave_expression(self, eps: float = 1e-6) -> cp.Expression:
        x = self.exponents
        if x.value is None:
            return None

        x = np_vec(x.value, order="F")
        y = cp.vec(self.weights, order="F")
        arg = y @ np.exp(x)
        return cp.log(arg)

    def get_convex_expression(self, eps: float = 1e-6) -> cp.Expression:
        x = self.weights
        if x.value is None:
            return None

        x = np_vec(x.value, order="F")
        y = cp.vec(self.exponents, order="F")

        nonneg = np.where(x > eps)[0]
        arg = y[nonneg] + np.log(x)[nonneg]
        return cp.log_sum_exp(arg)

    def _get_K_repr(self, local_to_glob: LocalToGlob, switched: bool = False) -> KRepresentation:
        f_local = cp.Variable(self.weights.size, name="f_wlse")
        t = cp.Variable(name="t_wlse")
        u = cp.Variable(name="u_wlse")

        epi_exp = cp.Variable(self.exponents.size, name="exp_epi")
        constraints = [
            epi_exp >= self.exponents,  # handles composition in exponent
            ExpCone(cp.reshape(epi_exp + u, (f_local.size,)), np.ones(f_local.size), f_local),
            t >= -u - 1,
        ]

        t_global = cp.Variable(name="t_global")
        if self.concave_composition:
            constraints += [t_global == t]
            f_global = cp.Variable(self.weights.size, name="f_global_wlse_comp")
            constraints += [f_global == f_local]
        else:
            B, c = affine_to_canon(self.weights, local_to_glob, switched)
            constraints += [t_global == t + f_local @ c]
            f_global = cp.Variable(
                local_to_glob.y_size if not switched else local_to_glob.x_size,
                name="f_global_wlse",
            )
            constraints += [f_global == B.T @ f_local]

        K_repr = KRepresentation(
            f=f_global,
            t=t_global,
            constraints=constraints,
        )

        switching_variables = self.exponents.variables()  # if not self.concave_composition else [z]
        precomp = (
            cp.Variable(self.weights.size, name="z_wlse") if self.concave_composition else None
        )
        K_out = (
            switch_convex_concave(
                constraints,
                f_global,
                t_global,
                switching_variables,
                local_to_glob,
                precomp,
            )
            if switched
            else K_repr
        )

        if self.concave_composition:
            if not switched:
                x_vars_1 = self.convex_variables()
                K_switch_1 = switch_convex_concave(
                    K_out.constraints,
                    K_out.f,
                    K_out.t,
                    x_vars_1,
                    local_to_glob,
                    precomp=precomp,
                )
                K_switch_1.constraints += [self.weights >= precomp]

                # dualize the outer concave exp variables if switched
                x_vars_2 = self.concave_variables()
                K_out = switch_convex_concave(
                    K_switch_1.constraints,
                    K_switch_1.f,
                    K_switch_1.t,
                    x_vars_2,
                    local_to_glob,
                )

            else:
                K_out.constraints += [precomp <= self.weights]

        if not self.weights.is_nonneg():
            if switched:
                K_out.constraints += [self.weights >= 0]
            else:
                K_out.y_constraints += [self.weights >= 0]

        concave_fun = (
            (lambda x: self.get_concave_expression())
            if not switched
            else (
                lambda x: -self.get_convex_expression()
                if self.get_convex_expression() is not None
                else None
            )
        )
        K_out.concave_expr = concave_fun

        return K_out

    def convex_variables(self) -> list[cp.Variable]:
        return self.exponents.variables()

    def concave_variables(self) -> list[cp.Variable]:
        return self.weights.variables()

    def is_incr(self, idx: int) -> bool:
        return True  # increasing in both arguments since y nonneg


class saddle_quad_form(SaddleAtom):
    def __init__(self, x: cp.Expression, P: cp.Expression) -> None:
        assert isinstance(x, cp.Expression)
        assert isinstance(P, cp.Expression)

        assert x.size == P.shape[0] == P.shape[1]

        self.x = x
        self.P = P

        super().__init__(x, P)

    def _numeric(self, values: list[np.ndarray]) -> np.ndarray:
        x, P = values
        return x.T @ P @ x

    def is_dsp(self) -> bool:
        return self.P.is_psd() and self.x.is_affine()

    def get_concave_expression(self) -> cp.Expression:
        x = self.x
        assert x.value is not None
        x = np_vec(x.value, order="F")
        P = self.P
        return x.T @ P @ x

    def get_convex_expression(self) -> cp.Expression:
        P = self.P.value
        assert P is not None

        x = self.x
        return cp.quad_form(x, P)

    def _get_K_repr(self, local_to_glob: LocalToGlob, switched: bool = False) -> KRepresentation:
        n = self.x.size
        F_local = cp.Variable((n, n), name="F_quad_form_local", PSD=True)
        A = cp.Variable((n + 1, n + 1), name="A_quad_form", PSD=True)
        t = cp.Variable(name="t_quad_form")

        constraints = [
            A[:n, :n] == F_local,
            A[-1, -1] == 1,
            A[:n, -1] == self.x,
            A[-1, :n] == self.x.T,
            t == 0,
        ]

        B, c = affine_to_canon(self.P, local_to_glob, switched)
        F_global = cp.Variable(
            local_to_glob.y_size if not switched else local_to_glob.x_size,
            name="f_global_saddle_quad_form",
        )
        constraints += [F_global == B.T @ cp.vec(F_local)]

        K_repr = KRepresentation(
            f=F_global,
            t=t,
            constraints=constraints,
            concave_expr=lambda x: self.get_concave_expression(),
        )

        if switched:
            K_repr = switch_convex_concave(
                constraints, F_global, t, self.convex_variables(), local_to_glob
            )
            K_repr.concave_expr = lambda x: self.get_convex_expression()

        return K_repr

    def name(self) -> str:
        return "saddle_quad_form(" + self.x.name() + ", " + self.P.name() + ")"

    def convex_variables(self) -> list[cp.Variable]:
        return self.x.variables()

    def concave_variables(self) -> list[cp.Variable]:
        return self.P.variables()

    def sign_from_args(self) -> tuple[bool, bool]:
        return (True, False)


class quasidef_quad_form(SaddleAtom):
    def __init__(
        self, x: cp.Expression, y: cp.Expression, P: cp.Constant, Q: cp.Constant, S: cp.Constant
    ) -> None:
        assert isinstance(x, cp.Expression)
        assert isinstance(y, cp.Expression)

        P, Q, S = map(cp.Expression.cast_to_const, (P, Q, S))

        assert isinstance(P, cp.Constant)
        assert isinstance(Q, cp.Constant)
        assert isinstance(S, cp.Constant)

        assert len(x.shape) == 1 and len(y.shape) == 1
        assert x.size == P.shape[0] == P.shape[1] == S.shape[0]
        assert y.size == Q.shape[0] == Q.shape[1] == S.shape[1]

        self.x = x
        self.y = y

        self.P = P
        self.Q = Q
        self.S = S

        super().__init__(x, y)

    def _numeric(self, values: list[np.ndarray]) -> np.ndarray:
        x, y = values
        return x.T @ self.P.value @ x + y.T @ self.Q.value @ y + 2 * x.T @ self.S.value @ y

    def is_dsp(self) -> bool:
        x_cvx = self.P.is_psd() and self.x.is_affine()
        y_ccv = self.Q.is_nsd() and self.y.is_affine()
        return x_cvx and y_ccv

    def get_concave_expression(self) -> cp.Expression:
        x = self.x
        assert x.value is not None
        x = np_vec(x.value, order="F")
        return x.T @ self.P @ x + cp.quad_form(self.y, self.Q) + 2 * x.T @ self.S @ self.y

    def get_convex_expression(self) -> cp.Expression:
        y = self.y
        assert y.value is not None
        y = np_vec(y.value, order="F")
        P = self.P
        return cp.quad_form(self.x, P) + y.T @ self.Q @ y + 2 * self.x.T @ self.S @ y

    def _get_K_repr(self, local_to_glob: LocalToGlob, switched: bool = False) -> KRepresentation:
        cvx_expr = cp.quad_form(self.x, self.P) if not switched else cp.quad_form(self.y, -self.Q)
        K_cvx = K_repr_ax(cvx_expr)

        ccv_expr = -cp.quad_form(self.y, -self.Q) if not switched else -cp.quad_form(self.x, self.P)
        K_ccv = K_repr_by(ccv_expr, local_to_glob)

        K_inner = (
            K_repr_bilin(self.x, 2 * self.S @ self.y, local_to_glob)
            if not switched
            else K_repr_bilin(self.y, 2 * self.S.T @ self.x, local_to_glob)
        )

        K_repr = KRepresentation.sum_of_K_reprs([K_cvx, K_ccv, K_inner])

        return K_repr

    def name(self) -> str:
        return (
            f"quasidef_quad_form({self.x.name()}, {self.y.name()}, {self.P.name()}, "
            f"{self.Q.name()}, {self.S.name()})"
        )

    def convex_variables(self) -> list[cp.Variable]:
        return self.x.variables()

    def concave_variables(self) -> list[cp.Variable]:
        return self.y.variables()


class weighted_norm2(SaddleAtom):
    def __init__(self, x: cp.Expression, y: cp.Expression) -> None:
        """
        Implements the function f(x,y) = (sum_i y_i x_i^2)^(1/2) for vectors x and y.
        The weights, y, must be non-negative. If they are non recognized by
        cvxpy as nonnegative, an implicit domain constraint is added and a
        warning is provided.
        The exponents can be any affine, or nonnegative convex cvxpy expression.

        The conic representation is:
            (sum_i y_i x_i^2)^(1/2)
            min. f, t       f @ y + t
            subject to      f >= 0
                            t >= 0
                            sqrt(t*f) >= 0.5 * abs(x)
        """

        assert isinstance(x, cp.Expression)
        assert isinstance(y, cp.Expression)

        if not y.is_nonneg():
            warnings.warn(
                "Weights are non-positive. The domain of weighted_norm2 is y >="
                " 0. The implicit constraint y >= 0 will be added to"
                " the problem."
            )

        self.concave_composition = not y.is_affine()

        self.x = x
        self.y = y

        assert len(x.shape) == 1 and len(y.shape) == 1
        assert x.shape == y.shape

        super().__init__(x, y)

    def _numeric(self, values: list[np.ndarray]) -> np.ndarray:
        x, y = values
        return np.sqrt(np.sum(y * x**2))

    def is_dsp(self) -> bool:
        x_cvx = cp.norm2(self.x).is_convex()
        y_ccv = self.y.is_concave()
        return x_cvx and y_ccv

    def _get_K_repr(self, local_to_glob: LocalToGlob, switched: bool) -> KRepresentation:
        f_local = cp.Variable(self.y.size, name="f_wnorm2")
        t = cp.Variable(name="t_wnorm2")

        # See https://www2.isye.gatech.edu/~nemirovs/LMCOLN2022Fall.pdf 2.3.5
        tau = cp.Variable(self.x.size, name="tau")
        sum_half = (f_local + t) / 2
        diff_half = (f_local - t) / 2
        stack = cp.vstack([tau, diff_half])

        constraints = [
            t >= 0,
            f_local >= 0,
            cp.norm2(stack, axis=0) <= sum_half,
            tau >= 0.5 * cp.abs(self.x),
        ]

        t_global = cp.Variable(name="t_global")
        if self.concave_composition:
            constraints += [t_global == t]
            f_global = cp.Variable(self.y.size, name="f_global_wnorm2_comp")
            constraints += [f_global == f_local]
        else:
            B, c = affine_to_canon(self.y, local_to_glob, switched)
            constraints += [t_global == t + f_local @ c]
            f_global = cp.Variable(
                local_to_glob.y_size if not switched else local_to_glob.x_size,
                name="f_global_wnorm2",
            )
            constraints += [f_global == B.T @ f_local]

        K_repr = KRepresentation(
            f=f_global,
            t=t_global,
            constraints=constraints,
        )

        switching_variables = self.x.variables()

        precomp = cp.Variable(self.y.size, name="z_wnorm2") if self.concave_composition else None
        K_out = (
            switch_convex_concave(
                constraints,
                f_global,
                t_global,
                switching_variables,
                local_to_glob,
                precomp,
            )
            if switched
            else K_repr
        )

        if self.concave_composition:
            if not switched:
                x_vars_1 = self.convex_variables()
                K_switch_1 = switch_convex_concave(
                    K_out.constraints,
                    K_out.f,
                    K_out.t,
                    x_vars_1,
                    local_to_glob,
                    precomp=precomp,
                )
                K_switch_1.constraints += [self.y >= precomp]

                # dualize the outer concave x variables if switched
                x_vars_2 = self.concave_variables()
                K_out = switch_convex_concave(
                    K_switch_1.constraints,
                    K_switch_1.f,
                    K_switch_1.t,
                    x_vars_2,
                    local_to_glob,
                )

            else:
                K_out.constraints += [precomp <= self.y]

        if not self.y.is_nonneg():
            if switched:
                K_out.constraints += [self.y >= 0]
            else:
                K_out.y_constraints += [self.y >= 0]

        concave_fun = (
            (lambda x: self.get_concave_expression())
            if not switched
            else (
                lambda x: -self.get_convex_expression()
                if self.get_convex_expression() is not None
                else None
            )
        )
        K_out.concave_expr = concave_fun

        return K_out

    def get_concave_expression(self) -> cp.Expression:
        x = self.x.value
        assert x is not None
        y = self.y

        return cp.sqrt(y.T @ np.square(x))

    def get_convex_expression(self) -> cp.Expression:
        y = self.y.value
        assert y is not None
        x = self.x

        return cp.norm2(cp.multiply(x, np.sqrt(y)))

    def convex_variables(self) -> list[cp.Variable]:
        return self.x.variables()

    def concave_variables(self) -> list[cp.Variable]:
        return self.y.variables()

    def sign_from_args(self) -> tuple[bool, bool]:
        return (True, False)

    def is_incr(self, idx: int) -> bool:
        return self.x.is_nonneg()



import cvxpy as cp
import numpy as np

from dsp.cone_transforms import LocalToGlob, affine_to_canon


def test_affine_to_canon():
    x = cp.Variable(2, name="x", nonneg=True)
    aff = cp.sum(x - 1)

    ltg = LocalToGlob([], [x])

    B, c = affine_to_canon(aff, ltg, False)
    assert np.allclose(B.toarray(), np.ones(2)[None, :])
    assert np.allclose(c, -2)



import cvxpy as cp
import numpy as np
import pytest

import dsp
from dsp import (
    conjugate,
    inner,
    saddle_inner,
    saddle_max,
    saddle_min,
    saddle_quad_form,
    weighted_log_sum_exp,
)
from dsp.cone_transforms import (
    K_repr_ax,
    K_repr_x_Gy,
    LocalToGlob,
    add_cone_constraints,
    get_cone_repr,
    minimax_to_min,
)
from dsp.local import LocalVariable, LocalVariableError
from dsp.parser import DSPError
from dsp.problem import (
    MinimizeMaximize,
    SaddlePointProblem,
    validate_all_saddle_extrema,
)


def test_matrix_game_x_Gy():
    n = 2

    x = cp.Variable(n, name="x")
    y = cp.Variable(n, name="y")
    yy = cp.Variable(n, name="yy")

    X_constraints = [
        -1 <= x,
        x <= 1,
    ]
    Y_constraints = [-1 <= yy, yy <= 1, yy <= y, y <= yy]

    F = y + 0.5

    ltg = LocalToGlob([x], [y, yy])

    K = K_repr_x_Gy(F, x, ltg)
    prob = cp.Problem(*minimax_to_min(K, X_constraints, Y_constraints, [y, yy], ltg))
    optimal_value = prob.solve(solver=cp.SCS)
    assert np.isclose(prob.value, 0, atol=1e-4)
    assert np.isclose(optimal_value, 0, atol=1e-4)


@pytest.mark.parametrize("a,expected", [(cp.exp, np.exp(2)), (cp.square, 4)])
def test_ax(a, expected):
    x = cp.Variable()
    y = cp.Variable()
    ax = a(x)

    X_const = [2 <= x, x <= 4]
    Y_const = [y == 0]

    K = K_repr_ax(ax)
    local_to_glob = LocalToGlob([x], [y])
    prob = cp.Problem(*minimax_to_min(K, X_const, Y_const, [y], local_to_glob))
    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, expected)


@pytest.mark.parametrize(
    "b_neg",
    [
        lambda y: cp.exp(y),
        lambda y: cp.inv_pos(y),
        (lambda y: cp.square(y)),
        (lambda y: cp.abs(y)),
        (lambda y: -cp.log(y)),
    ],
)
@pytest.mark.parametrize("y_val", range(1, 4))
def test_by(b_neg, y_val):
    expected = -b_neg(y_val).value

    x = cp.Variable(name="x")
    y = cp.Variable(name="y")

    by = -b_neg(y)

    if by.domain and y_val <= 0:
        return  # skip test

    X_const = [x == 0]
    Y_const = [y == y_val]

    prob = SaddlePointProblem(MinimizeMaximize(by), X_const + Y_const, minimization_vars={x})
    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, expected, atol=1e-3)
    assert np.isclose(y.value, y_val)


@pytest.mark.parametrize("y_val", range(-3, 3))
def test_epigraph_exp(y_val):
    y = cp.Variable(name="y")
    b = cp.square(cp.square(y))

    t_primal = cp.Variable(name="t_by_primal")

    constraints = [
        t_primal >= b,
    ]

    (
        var_to_mat_mapping,
        s_bar,
        cone_dims,
    ) = get_cone_repr(constraints, [y, t_primal])

    R_bar = var_to_mat_mapping[y.id]
    p_bar = var_to_mat_mapping[t_primal.id]
    Q_bar = var_to_mat_mapping["eta"]

    s_bar = s_bar.reshape(-1, 1)
    u_temp = cp.Variable((Q_bar.shape[1], 1))

    Ax_b = s_bar - (R_bar * y + t_primal * p_bar + Q_bar @ u_temp)

    cone_constraints, _ = add_cone_constraints(Ax_b, cone_dims, dual=False)

    prob = cp.Problem(
        cp.Minimize(t_primal),
        [*cone_constraints, y == y_val],
    )
    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, y_val**4, atol=1e-3)

    u = cp.Variable((Q_bar.shape[0], 1))
    cone_constraints, u = add_cone_constraints(u, cone_dims, dual=True)
    max_prob = cp.Problem(
        cp.Maximize((R_bar.T @ u) * y_val - s_bar.T @ u),
        [-u.T @ p_bar == 1, Q_bar.T @ u == 0] + cone_constraints,
    )
    max_prob.solve(solver=cp.SCS)
    assert max_prob.status == cp.OPTIMAL
    assert np.isclose(max_prob.value, y_val**4, atol=1e-7)


def test_matrix_game_nemirovski_Fx_Gy():
    n = 2

    x = cp.Variable(n, name="x")
    y = cp.Variable(n, name="y")

    F_x = x + 0.5
    G_y = y + 0.5

    FxGy = inner(F_x, G_y)

    objective = MinimizeMaximize(FxGy)
    constraints = [-1 <= x, x <= 1, -1 <= y, y <= 1]
    prob = SaddlePointProblem(objective, constraints)
    prob.solve(solver=cp.SCS)
    assert np.isclose(prob.value, 0, atol=1e-4)
    assert np.isclose(prob.value, FxGy.value, atol=1e-4)


def test_overload_bilin():
    x = cp.Variable()
    y = cp.Variable()

    objective = MinimizeMaximize(x * y)
    constraints = [-1 <= x, x <= 1, -1.2 <= y, y <= -0.8]
    prob = SaddlePointProblem(objective, constraints, minimization_vars={x}, maximization_vars={y})

    with pytest.raises(DSPError, match="Use inner instead"):
        prob.solve()


@pytest.mark.parametrize(
    "obj",
    [lambda x, y: inner(x, 1 + y), lambda x, y: x + inner(x, y)],
)
def test_saddle_composition(obj):
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")

    objective = MinimizeMaximize(obj(x, y))

    constraints = [-1 <= x, x <= 1, -1.2 <= y, y <= -0.8]
    prob = SaddlePointProblem(objective, constraints, minimization_vars={x}, maximization_vars={y})

    prob.solve(solver=cp.SCS)
    assert prob.status == cp.OPTIMAL
    assert np.isclose(prob.value, 0)
    assert np.isclose(y.value, -1)


def test_constant():
    obj = MinimizeMaximize(10)
    problem = SaddlePointProblem(obj)
    problem.solve(solver=cp.SCS)
    assert problem.value == 10


def test_variable():
    k = cp.Variable(name="k")
    constraints = [-10 <= k, k <= 10]

    obj = MinimizeMaximize(k)
    with pytest.raises(DSPError, match="Cannot split"):
        SaddlePointProblem(obj, constraints).solve()

    with pytest.raises(DSPError, match="Cannot resolve"):
        SaddlePointProblem(obj, []).solve()

    problem = SaddlePointProblem(obj, constraints, minimization_vars={k})
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, -10)

    problem = SaddlePointProblem(obj, constraints, maximization_vars={k})
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, 10)

    # No need to specify when convex/concave terms are present
    obj = MinimizeMaximize(k + 1e-10 * cp.pos(k))
    problem = SaddlePointProblem(obj, constraints)
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, -10)

    obj = MinimizeMaximize(k - 1e-10 * cp.pos(k))
    problem = SaddlePointProblem(obj, constraints)
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, 10)


def test_sum():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")

    obj = MinimizeMaximize(x + y)
    constraints = [-1 <= x, x <= 1, -2 <= y, y <= 2]
    problem = SaddlePointProblem(obj, constraints, minimization_vars={x}, maximization_vars={y})
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, 1.0)


def test_mixed_curvature_affine():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")

    obj = MinimizeMaximize(cp.exp(x) + cp.log(y) + np.array([1, 2]) @ cp.vstack([x, y]))

    constraints = [x == 0, y == 1]
    problem = SaddlePointProblem(obj, constraints)

    problem.solve()
    assert np.isclose(problem.value, 3.0)

    obj = MinimizeMaximize(cp.exp(x) + cp.log(y) + x + 2 * y)
    problem = SaddlePointProblem(obj, constraints)
    problem.solve()

    assert np.isclose(problem.value, 3)


def test_indeterminate_problem():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")
    z = cp.Variable(name="z")
    obj = MinimizeMaximize(cp.exp(x) + cp.log(y) + z)
    with pytest.raises(DSPError, match="Cannot resolve"):
        SaddlePointProblem(obj).solve()

    prob = SaddlePointProblem(obj, minimization_vars={z}, constraints=[y >= 0, x >= 0, z >= 0])
    assert set(prob.x_prob.variables()) & set(prob.variables()) == {x, z}
    assert set(prob.y_prob.variables()) & set(prob.variables()) == {y}


def test_stacked_variables():
    x1 = cp.Variable(2)
    x2 = cp.Variable()
    y1 = cp.Variable()
    y2 = cp.Variable(2)

    obj = MinimizeMaximize(cp.sum(cp.exp(x1)) + cp.square(x2) + cp.log(y1) + cp.sum(cp.sqrt(y2)))
    constraints = [
        -1 <= x1,
        x1 <= 1,
        -2 <= x2,
        x2 <= 2,
        1 <= y1,
        y1 <= 3,
        1 <= y2,
        y2 <= 3,
    ]
    problem = SaddlePointProblem(obj, constraints)
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, 2 * np.exp(-1) + np.log(3) + 2 * np.sqrt(3))


def test_Gx_Fy():
    x = cp.Variable(2, name="x")
    y = cp.Variable(2, name="y")

    with pytest.warns(UserWarning, match="is non-positive"):
        f = saddle_inner(cp.exp(x), cp.log(y))
    obj = MinimizeMaximize(f)
    assert f.is_nonneg()
    assert f.is_incr(0)
    assert f.is_incr(1)

    low = 2
    constraints = [low <= x, x <= 4, 2 <= y, y <= 3]

    problem = SaddlePointProblem(obj, constraints)
    problem.solve(solver=cp.SCS)

    assert np.isclose(problem.value, 2 * np.exp(low) * np.log(3))


def test_robust_constraint():
    x = cp.Variable(name="x")
    y = LocalVariable(name="y", nonneg=True)
    assert repr(y).startswith("LocalVariable")

    obj = MinimizeMaximize(cp.square(x))

    constraints = [x >= 1]

    constraints += [saddle_max(weighted_log_sum_exp(x, y), [y <= 1]) <= 1]

    invalid_saddle_problem = SaddlePointProblem(
        obj, constraints, minimization_vars=[x], maximization_vars=[y]
    )
    assert not invalid_saddle_problem.is_dsp()
    assert not dsp.is_dsp(invalid_saddle_problem)
    with pytest.raises(DSPError, match="Likely passed unused variables"):
        invalid_saddle_problem.solve()

    problem = SaddlePointProblem(obj, constraints)
    assert dsp.is_dsp(problem)
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, 1.0, atol=1e-4)


def test_robust_constraint_min():
    x = cp.Variable(name="x")
    y = LocalVariable(name="y", nonneg=True)

    obj = dsp.MinimizeMaximize(x)

    constraints = [x >= 1]

    constraints += [saddle_max(weighted_log_sum_exp(x, y), [y <= 1]) <= 1]

    problem = SaddlePointProblem(obj, constraints, minimization_vars=[x])
    problem.solve(solver=cp.SCS)
    assert np.isclose(problem.value, 1.0, atol=1e-4)

    obj = cp.Minimize(x)
    problem = cp.Problem(obj, constraints)
    problem.solve(solver=cp.SCS)
    validate_all_saddle_extrema(problem)
    assert np.isclose(problem.value, 1.0, atol=1e-4)


def test_robust_constraint_inf():
    """
    Test that we can handle robust constraint inf_x f(x,y) >= eta.
    """
    x = LocalVariable(name="x_dummy")
    y = cp.Variable(name="y", nonneg=True)

    x_val = 1.0
    y_val = 2.0

    saddle_min_expr = saddle_min(weighted_log_sum_exp(x, y), [x >= x_val])

    constraints = [y <= y_val, saddle_min_expr >= np.log(y_val * np.exp(x_val))]

    obj = cp.Maximize(y)

    problem = cp.Problem(obj, constraints)
    problem.solve(solver=cp.SCS)
    validate_all_saddle_extrema(problem)
    assert np.isclose(problem.value, y_val, atol=1e-4)
    assert set(saddle_min_expr.concave_variables()) == {y}
    assert set(saddle_min_expr.convex_variables()) == {x}
    assert np.isclose(saddle_min_expr.value, np.log(y_val * np.exp(x_val)), atol=1e-4)
    assert (
        saddle_min_expr.name() == "saddle_min(weighted_log_sum_exp(x_dummy, y), [1.0 <= x_dummy])"
    )


@pytest.mark.parametrize("n", [1, 2, 3])
@pytest.mark.parametrize(
    "expr,expected",
    [
        (lambda Y: cp.log_det(Y), lambda n: n),
        (lambda Y: cp.trace(Y), lambda n: n * np.e),
        (lambda Y: -cp.lambda_max(Y), lambda n: -np.e),
    ],
)
def test_PSD_saddle(n, expr, expected):
    Y = cp.Variable((n, n), PSD=True, name="Y")
    f = expr(Y)

    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [Y == np.e * np.eye(n)], maximization_vars=[Y])

    prob.solve(solver=cp.SCS)
    assert np.isclose(prob.value, expected(n), atol=1e-4)


def test_psd_exp():
    n = 2
    Y = cp.Variable((n, n), PSD=True, name="Y")

    obj = MinimizeMaximize(cp.exp(cp.lambda_max(Y)))

    prob = SaddlePointProblem(obj, [Y >> np.eye(n)], minimization_vars=[Y])

    prob.solve()

    assert np.isclose(prob.value, np.e, atol=1e-4)


def test_worst_case_covariance():
    kappa = 0.5
    Sigma = np.array([[1, -0.3], [-0.3, 2]])
    delta = LocalVariable((2, 2), symmetric=True, name="delta")
    Sigma_pert = LocalVariable((2, 2), PSD=True, name="Sigma_pert")
    v = cp.Variable(2, name="v")
    obj = saddle_quad_form(v, Sigma_pert)
    constraints = [
        Sigma + delta == Sigma_pert,
        cp.abs(delta) <= kappa * np.sqrt(np.outer(Sigma.diagonal(), Sigma.diagonal())),
    ]

    worst_case_risk = saddle_max(obj, constraints)

    v_val = np.array([0.70348158, 0.29651842])  # This is the optimal value obtained via closed form

    wc_ref = v_val.T @ Sigma @ v_val + kappa * (np.sqrt(np.diag(Sigma)) @ np.abs(v_val)) ** 2

    v.value = v_val
    assert np.isclose(wc_ref, worst_case_risk.value, atol=1e-4)
    assert worst_case_risk.name().startswith("saddle_max(saddle_quad_form(")

    # TODO: solver does not find the optimal value without fixing the values to opt
    prob = cp.Problem(cp.Minimize(worst_case_risk), [cp.sum(v) == 1, v >= 0, v == v_val])

    prob.solve(solver=cp.SCS)

    wc_ob_ref = cp.quad_form(v, Sigma) + kappa * cp.square(np.sqrt(np.diag(Sigma)) @ cp.abs(v))
    ref_prob = cp.Problem(cp.Minimize(wc_ob_ref), [cp.sum(v) == 1, v >= 0])
    ref_prob.solve(solver=cp.SCS)

    assert np.isclose(ref_prob.value, prob.value, atol=1e-4)


def test_aux_variable_in_constraints():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")
    z = cp.Variable(name="z")

    with pytest.warns(UserWarning, match="Weights are non-positive"):
        wlse = weighted_log_sum_exp(x, y)

    saddle_prob = SaddlePointProblem(MinimizeMaximize(wlse), [x == 1, y <= z, z <= 1])
    assert saddle_prob.is_dsp()
    saddle_prob.solve()
    assert np.isclose(saddle_prob.value, 1, atol=1e-4)


def test_SE_variable_in_constraint():
    x = cp.Variable(name="x")
    y = LocalVariable(name="y")
    z = LocalVariable(name="z")
    z_nonlocal = cp.Variable(name="z_nonloacl")

    with pytest.warns(UserWarning, match="Weights are non-positive"):
        wlse = weighted_log_sum_exp(x, y)

    with pytest.raises(LocalVariableError):
        saddle_max(wlse, [y <= z_nonlocal, z_nonlocal <= 1])

    se = saddle_max(wlse, [y <= z, z <= 1])
    problem = cp.Problem(cp.Minimize(se), [x == 1])
    problem.solve()
    validate_all_saddle_extrema(problem)
    assert np.isclose(problem.value, 1, atol=1e-4)


def test_unconstrained():
    x = cp.Variable(1, name="x")
    y = cp.Variable(1, name="y")
    f = cp.square(x) - cp.square(y)
    obj = MinimizeMaximize(f)

    saddle_problem = SaddlePointProblem(obj)
    saddle_problem.is_dsp()

    saddle_problem.solve()
    assert np.isclose(saddle_problem.value, 0)


def test_partially_constrained():
    x = cp.Variable(2, name="x")
    y1 = cp.Variable(2, name="y1")
    y2 = cp.Variable(1, name="y2")
    f = inner(x, y1) - cp.square(y2)
    obj = MinimizeMaximize(f)

    saddle_problem = SaddlePointProblem(obj, [x == 1, y1 == 1])
    saddle_problem.is_dsp()

    saddle_problem.solve()
    assert np.isclose(saddle_problem.value, 2)


def test_saddle_inner_neg():
    x = cp.Variable(1, name="x")
    y = cp.Variable(1, name="y", nonneg=True)
    f = saddle_inner(cp.square(x) - 1, y)
    obj = MinimizeMaximize(f)

    assert not obj.is_dsp()

    saddle_problem = SaddlePointProblem(obj)
    assert not saddle_problem.is_dsp()

    with pytest.raises(DSPError):
        saddle_problem.solve()


def test_scalar():
    x = cp.Variable()
    y = cp.Variable()
    f = inner(x, y)
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [-1 <= x, x <= 1, -1 <= y, y <= 1])
    prob.solve()
    assert np.isclose(prob.value, 0)
    assert np.isclose(x.value, 0)
    assert np.isclose(y.value, 0)


def test_inner_properties():

    x_nonneg = cp.Variable(nonneg=True)
    x_nonpos = cp.Variable(nonpos=True)

    y = cp.Variable()
    y_nonneg = cp.Variable(nonneg=True)
    y_nonpos = cp.Variable(nonpos=True)

    assert inner(x_nonneg, y).is_incr(1)
    assert inner(x_nonpos, y).is_decr(1)

    assert inner(x_nonneg, y_nonneg).is_nonneg()
    assert inner(x_nonneg, y_nonpos).is_nonpos()


def test_non_local_variable():
    x = cp.Variable(1, name="x")
    y = cp.Variable(1, name="y")

    with pytest.raises(LocalVariableError):
        conjugate(inner(x, y))


def test_pure_convex():
    x = cp.Variable(name="x")
    f = cp.square(x)
    obj = MinimizeMaximize(f)

    assert obj.is_dsp()

    saddle_problem = SaddlePointProblem(obj)
    assert saddle_problem.is_dsp()

    saddle_problem.solve()
    assert np.isclose(saddle_problem.value, 0)


def test_pure_concave():
    y = cp.Variable(name="y")
    f = -cp.square(y)
    obj = MinimizeMaximize(f)

    assert obj.is_dsp()

    saddle_problem = SaddlePointProblem(obj)
    assert saddle_problem.is_dsp()

    saddle_problem.solve()
    assert np.isclose(saddle_problem.value, 0)


def test_local_variable_setter():
    x = LocalVariable(name="x")
    y = cp.Variable(name="y")
    se = saddle_min(y + x, [x <= 1])
    x.value = 1
    assert x.value == 1  # currently value only computes



import cvxpy as cp
import numpy as np
import pytest

import dsp
from dsp import weighted_norm2


@pytest.mark.parametrize("n", [1, 2, 3, 4, 5])
def test_weighted_norm2(n):
    x = cp.Variable(n, name="x", nonneg=True)
    y = cp.Variable(n, name="y", nonneg=True)

    norm = weighted_norm2(x, y)
    assert norm.value is None
    assert not norm.is_decr(0)
    assert norm.is_dsp()

    obj = dsp.MinimizeMaximize(norm)
    constraints = [y <= 2, x >= 1]
    prob = dsp.SaddlePointProblem(obj, constraints)
    prob.solve()

    y_expected = np.ones(n) * 2
    x_expected = np.ones(n) * 1
    expected_prob_val = np.sqrt(y_expected @ x_expected**2)

    assert prob.status == "optimal"
    assert np.isclose(prob.value, expected_prob_val)
    assert np.allclose(x.value, x_expected)
    assert np.allclose(y.value, y_expected)

    y.value = y_expected
    min_val = cp.Problem(cp.Minimize(norm.get_convex_expression()), constraints).solve()

    x.value = x_expected
    max_val = cp.Problem(cp.Maximize(norm.get_concave_expression()), constraints).solve()

    assert np.isclose(min_val, expected_prob_val)
    assert np.isclose(max_val, expected_prob_val)


@pytest.mark.parametrize("n", [1, 2, 3, 4, 5])
def test_neg_weighted_norm2(n: int) -> None:
    x = cp.Variable(n, name="x", nonneg=True)
    y = cp.Variable(n, name="y", nonneg=True)

    neg_norm = -weighted_norm2(y, x)
    assert neg_norm.is_dsp()

    obj = dsp.MinimizeMaximize(neg_norm)
    prob = dsp.SaddlePointProblem(obj, [y >= 2, x <= 1])
    prob.solve()

    y_expected = np.ones(n) * 2
    x_expected = np.ones(n) * 1

    assert prob.status == "optimal"
    assert np.isclose(prob.value, -np.sqrt(x_expected @ y_expected**2))
    assert np.allclose(x.value, x_expected)
    assert np.allclose(y.value, y_expected)


@pytest.mark.parametrize("n", [1, 2, 3, 4, 5])
def test_affine_precompositions(n: int) -> None:
    x = cp.Variable(n, name="x", nonneg=True)
    y = cp.Variable(n, name="y", nonneg=True)

    norm = weighted_norm2(2 * x + 1, 2 * y + 1)
    assert norm.is_dsp()

    obj = dsp.MinimizeMaximize(norm)
    prob = dsp.SaddlePointProblem(obj, [y <= 2, x >= 1])
    prob.solve()

    y_expected = np.ones(n) * 2
    x_expected = np.ones(n) * 1

    assert prob.status == "optimal"
    assert np.isclose(prob.value, np.sqrt((2 * y_expected + 1) @ (2 * x_expected + 1) ** 2))
    assert np.allclose(x.value, x_expected)
    assert np.allclose(y.value, y_expected)


@pytest.mark.parametrize("x_val", [-1, 0.1, 1])
@pytest.mark.parametrize(
    "x_expr", [lambda x: cp.square(x), lambda x: cp.square(x) - 1, lambda x: cp.exp(x)]
)
@pytest.mark.parametrize("n", [1, 2, 3, 4, 5])
def test_nonaffine_precompositions(x_val, x_expr, n: int) -> None:

    x = cp.Variable(n, name="x")
    y = cp.Variable(n, name="y")

    cvx_expr = x_expr(x)

    with pytest.warns(UserWarning, match="Weights are non-positive."):
        norm = weighted_norm2(cvx_expr, y)

    if cvx_expr.is_affine() or cvx_expr.is_nonneg():
        assert norm.is_dsp()

        obj = dsp.MinimizeMaximize(norm)

        x_range_constr = x >= x_val if x_val > 0 or cvx_expr.is_incr(0) else x <= x_val
        prob = dsp.SaddlePointProblem(obj, [y <= 2, x_range_constr])
        prob.solve()

        y_expected = np.ones(n) * 2
        x_expected = np.ones(n) * x_val

        assert prob.status == "optimal"
        assert np.isclose(prob.value, np.sqrt(y_expected @ (x_expr(x_expected).value ** 2)))
        assert np.allclose(x.value, x_expected)
        assert np.allclose(y.value, y_expected, atol=1e-5)
    else:
        assert not norm.is_dsp()


def test_concave_composition() -> None:
    n = 2
    x = cp.Variable(n, name="x")
    y = cp.Variable(n, name="y", nonneg=True)

    Gy = cp.log1p(y)
    norm = weighted_norm2(x, Gy)

    obj = dsp.MinimizeMaximize(norm)

    prob = dsp.SaddlePointProblem(obj, [y <= 2, x == 1])
    prob.solve()

    y_expected = np.ones(n) * 2
    x_expected = np.ones(n)

    assert prob.status == "optimal"
    assert np.isclose(prob.value, np.sqrt(np.log(1 + y_expected) @ (x_expected**2)))
    assert np.allclose(x.value, x_expected)
    assert np.allclose(y.value, y_expected, atol=1e-5)


def test_value() -> None:
    n = 2
    x = cp.Variable(n)
    y = cp.Variable(n)

    with pytest.warns(UserWarning, match="Weights are non-positive"):
        norm2 = weighted_norm2(x, y)

    assert norm2.value is None
    assert norm2.is_nonneg()
    assert not norm2.is_incr(0)
    assert not norm2.is_incr(1)

    x.value = np.arange(n)
    y.value = np.arange(n)

    assert norm2.value == np.sqrt(y.value @ np.square(x.value))



import cvxpy as cp
import numpy as np
import pytest

import dsp
from dsp import (
    LocalVariable,
    MinimizeMaximize,
    SaddlePointProblem,
    saddle_inner,
    saddle_min,
    saddle_quad_form,
)


@pytest.mark.skipif(cp.MOSEK not in cp.installed_solvers(), reason="MOSEK not installed")
def test_robust_bond():
    C = np.loadtxt("tests/example_data/robust_bond_portfolio/C.csv")
    p = np.loadtxt("tests/example_data/robust_bond_portfolio/p.csv")
    w_bar = np.loadtxt("tests/example_data/robust_bond_portfolio/target_weights.csv")
    y_nom = np.loadtxt("tests/example_data/robust_bond_portfolio/y_nominal.csv")
    h_mkt = (w_bar * 100) / p

    # Constants and parameters
    n, T = C.shape
    delta_max, kappa, omega = 0.02, 0.9, 1e-6
    B = 100
    V_lim = 90

    # Creating variables
    h = cp.Variable(n, nonneg=True)

    delta = LocalVariable(T)
    y = y_nom + delta

    # Objective
    phi = 0.5 * cp.norm1(cp.multiply(h, p) - cp.multiply(h_mkt, p))

    # Creating saddle min function
    V = 0
    for i in range(n):
        t_plus_1 = np.arange(T) + 1  # Account for zero-indexing
        V += saddle_inner(cp.exp(cp.multiply(-t_plus_1, y)), h[i] * C[i])

    Y = [
        cp.norm_inf(delta) <= delta_max,
        cp.norm1(delta) <= kappa,
        cp.sum_squares(delta[1:] - delta[:-1]) <= omega,
    ]

    V_wc = saddle_min(V, Y)

    # Creating and solving the problem
    problem = cp.Problem(cp.Minimize(phi), [h @ p == B, V_wc >= V_lim])
    problem.solve()  # 15.32

    assert problem.status == cp.OPTIMAL

    plotting = False

    if plotting:
        import matplotlib.pyplot as plt
        import pandas as pd

        # added tentative plotting, please review
        inds = (
            ((C > 0).astype(int) * np.arange(C.shape[1])).argmax(axis=1).argsort()
        )  # sort by maturity
        df = pd.DataFrame({"h": h.value[inds], "h_mkt": (h_mkt)[inds]})
        df.plot(kind="bar")
        plt.xlabel("Bond index (increasing maturity)")
        plt.ylabel("Holdings")
        plt.savefig("tests/example_data/robust_bond.pdf")
        # plt.show()

        df = pd.DataFrame({"y_nom": y_nom, "y_wc": y.value})
        df.plot()
        plt.xlabel(r"$t$")
        plt.ylabel("yield")
        plt.savefig("tests/example_data/yield.pdf")
        # plt.show()


def test_robust_markowitz():
    returns = np.loadtxt("tests/example_data/robust_portfolio_selection/ff_data.csv", delimiter=",")

    mu = np.mean(returns, axis=0)
    Sigma = np.cov(returns, rowvar=False)

    # Constants and parameters
    n = len(mu)
    rho, eta, gamma = 0.2, 0.2, 1

    # Creating variables
    w = cp.Variable(n, nonneg=True)

    delta_loc = LocalVariable(n)
    Sigma_perturbed = LocalVariable((n, n), PSD=True)
    Delta_loc = LocalVariable((n, n))

    # Creating saddle min function
    f = w @ mu + saddle_inner(delta_loc, w) - gamma * saddle_quad_form(w, Sigma_perturbed)

    Sigma_diag = Sigma.diagonal()
    local_constraints = [
        cp.abs(delta_loc) <= rho,
        Sigma_perturbed == Sigma + Delta_loc,
        cp.abs(Delta_loc) <= eta * np.sqrt(np.outer(Sigma_diag, Sigma_diag)),
    ]

    G = saddle_min(f, local_constraints)

    # Creating and solving the problem
    problem = cp.Problem(cp.Maximize(G), [cp.sum(w) == 1])
    problem.solve(solver=cp.SCS)  # 0.076

    nominal_objective = cp.Maximize(w @ mu - gamma * cp.quad_form(w, Sigma))

    robust_wc_utility = G.value
    robust_non_wc_utility = nominal_objective.value

    assert problem.status == cp.OPTIMAL

    # Creating and solving the problem without robustness
    problem = cp.Problem(nominal_objective, [cp.sum(w) == 1])
    problem.solve(solver=cp.SCS)  # 0.065
    assert problem.status == cp.OPTIMAL

    nominal_wc_utility = G.value
    nominal_non_wc_utility = nominal_objective.value

    assert robust_wc_utility > nominal_wc_utility
    assert robust_non_wc_utility < nominal_non_wc_utility


def test_arbitrage():
    p = np.ones(2)

    V = np.array([[1.5, 0.5], [0.5, 1.5]])

    x = cp.Variable(2, name="x")

    # No arbitrage
    obj = cp.Minimize(p @ x)
    constraints = [V @ x >= 0]
    no_arb_problem = cp.Problem(obj, constraints)
    no_arb_problem.solve()
    assert np.isclose(no_arb_problem.value, 0)

    # Arbitrage
    p[0] = 0.1  # arb possible for p[0] < 1/3
    obj = cp.Minimize(p @ x)
    constraints = [V @ x >= 0]
    arb_problem = cp.Problem(obj, constraints)
    arb_problem.solve()
    assert arb_problem.status == cp.UNBOUNDED

    # Robust arbitrage - small perturbation
    p_tilde = cp.Variable(2, name="p_tilde")
    obj = dsp.MinimizeMaximize(dsp.inner(x, p_tilde))
    constraints = [V @ x >= 0, cp.abs(p_tilde - p) <= 0.1]
    saddle_problem = dsp.SaddlePointProblem(obj, constraints)
    with pytest.raises(AssertionError):
        saddle_problem.solve()

    # Robust arbitrage - large perturbation
    constraints = [V @ x >= 0, cp.abs(p_tilde - p) <= 1]
    saddle_problem = dsp.SaddlePointProblem(obj, constraints)
    saddle_problem.solve()
    assert saddle_problem.status == cp.OPTIMAL
    assert np.isclose(saddle_problem.value, 0)


@pytest.mark.skipif(cp.MOSEK not in cp.installed_solvers(), reason="MOSEK not installed")
def test_robust_model_fitting():
    # Load data
    data = np.loadtxt("tests/example_data/robust_model_fitting/data.csv", delimiter=",", skiprows=1)

    p = data[:, 0]
    p = (p - np.mean(p)) / np.std(p)

    A = data[:, 1:]
    A = (A - np.mean(A, axis=0)) / np.std(A, axis=0)
    A = np.hstack((np.ones((len(p), 1)), A))

    # Constants
    m, n = A.shape
    k = 0.2 * m
    eta = 0.1

    # Creating variables
    theta = cp.Variable(n)
    weights = cp.Variable(m, nonneg=True)

    # Defining the loss function and the weight constraints
    loss = cp.square(A @ theta - p)
    regularizer = eta * cp.norm1(theta)
    objective = MinimizeMaximize(saddle_inner(loss, weights) + regularizer)
    constraints = [cp.sum(weights) == k, weights <= 1]

    # Creating and solving the problem
    problem = SaddlePointProblem(objective, constraints)
    problem.solve()  # XXX
    assert problem.status == cp.OPTIMAL

    robust_obj_robust_weights = problem.value

    # OLS problem
    ols_objective = cp.Minimize(cp.sum_squares(A @ theta - p) + regularizer)

    ols_obj_robust_weights = ols_objective.value

    problem = cp.Problem(ols_objective)
    problem.solve()
    assert problem.status == cp.OPTIMAL

    ols_obj_ols_weights = problem.value
    robust_obj_ols_weights = np.sum(np.sort(np.square(A @ theta.value - p))[-int(k) :])

    assert ols_obj_ols_weights < ols_obj_robust_weights
    assert robust_obj_robust_weights < robust_obj_ols_weights

    # Using sum_largest
    loss = cp.sum_largest(cp.square(A @ theta - p), k)
    objective = cp.Minimize(loss + regularizer)
    problem = cp.Problem(objective)
    problem.solve()
    assert problem.status == cp.OPTIMAL

    assert np.isclose(problem.value, robust_obj_robust_weights)


def test_svm():
    import pandas as pd

    one_hot = True  # Use one-hot encoding for pclass and 5 age bins
    intercept = False and one_hot  # Include intercept in model
    bins = 3  # Number of age bins
    train_port = ["Q"]  # C, Q, S: the port to use for training
    without_train = True  # Dont include training data in eval

    df = pd.read_csv("tests/example_data/robust_model_fitting/titanic.csv")

    df["sex"] = df["sex"] == "male"

    df["intercept"] = 1

    features = ["pclass", "sex", "age"] + (["intercept"] if intercept else [])

    df = df.dropna(subset=features)

    class_hot = pd.get_dummies(df["pclass"], prefix="pclass")

    age_bins = pd.get_dummies(pd.cut(df["age"], bins), prefix="age")

    df = pd.concat([df, class_hot, age_bins], axis=1)

    df["survived_bool"] = df["survived"].copy()
    df["survived"] = 2 * df["survived"] - 1

    if one_hot:
        features = ["sex"] + age_bins.columns.tolist() + class_hot.columns.tolist()

    df_short = df[df.embarked.isin(train_port)]

    y_train = df_short["survived"].values.astype(float)
    A_train = df_short[features].values.astype(float)

    surv = df_short["survived_bool"].values.astype(float)

    # Constants and parameters
    m, n = A_train.shape
    inds_0 = surv == 0
    inds_1 = surv == 1
    eta = 0.05

    # Creating variables
    theta = cp.Variable(n)
    beta_0 = cp.Variable()
    weights = cp.Variable(m, nonneg=True)
    surv_weight_0 = cp.Variable()
    surv_weight_1 = cp.Variable()

    # Defining the loss function and the weight constraints
    y_hat = A_train @ theta + beta_0
    loss = cp.pos(1 - cp.multiply(y_train, y_hat))
    objective = MinimizeMaximize(saddle_inner(loss, weights) + eta * cp.sum_squares(theta))

    constraints = [
        cp.sum(weights) == 1,
        0.408 - 0.05 <= weights @ surv,
        weights @ surv <= 0.408 + 0.05,
        weights[inds_0] == surv_weight_0,
        weights[inds_1] == surv_weight_1,
    ]

    # Creating and solving the problem
    problem = SaddlePointProblem(objective, constraints)
    problem.solve()

    assert problem.status == cp.OPTIMAL
    robust_theta = theta.value

    # nominal SVM problem
    const_weights = np.ones(m)
    problem = cp.Problem(
        cp.Minimize(cp.sum(cp.multiply(loss, const_weights)) + eta * cp.sum_squares(theta))
    )
    problem.solve()
    assert problem.status == cp.OPTIMAL
    nominal_theta = theta.value

    # Full sample
    if without_train:
        y = df[~df.embarked.isin(train_port)]["survived"].values.astype(float)
        A = df[~df.embarked.isin(train_port)][features].values.astype(float)
    else:
        y = df["survived"].values.astype(float)
        A = df[features].values.astype(float)

    print_results = False
    if print_results:
        print("Train accuracy nom.: ", accuracy(A_train @ nominal_theta, y_train))
        print("Train accuracy rob.: ", accuracy(A_train @ robust_theta, y_train))
        print("Test accuracy nom.: ", accuracy(A @ nominal_theta, y))
        print("Test accuracy rob.: ", accuracy(A @ robust_theta, y))

        print("-" * 80)

        print("Train loss nom.: ", avg_svm_loss_numpy(A_train @ nominal_theta, y_train))
        print("Train loss rob.: ", avg_svm_loss_numpy(A_train @ robust_theta, y_train))
        print("Test loss nom.: ", avg_svm_loss_numpy(A @ nominal_theta, y))
        print("Test loss rob.: ", avg_svm_loss_numpy(A @ robust_theta, y))

        print("-" * 80)

        print(
            "\n".join(
                ["robust theta"]
                + [
                    f"{name:20s}: {val:>10.4f}"
                    for name, val in zip(df_short[features].columns, robust_theta)
                ]
            )
        )
        print()
        print(
            "\n".join(
                ["nom theta"]
                + [
                    f"{name:20s}: {val:>10.4f}"
                    for name, val in zip(df_short[features].columns, nominal_theta)
                ]
            )
        )


def accuracy(scores, labels):
    scores[scores > 0] = 1
    scores[scores <= -0] = -1
    return np.mean(scores == labels)


def avg_svm_loss_numpy(y_hat, y):
    losses = np.maximum(0, 1 - y * y_hat)
    return np.mean(losses)



from __future__ import annotations

from typing import Iterable

import cvxpy as cp
import numpy as np

import dsp


class LocalVariableError(Exception):
    pass


class LocalVariable(cp.Variable):
    def __init__(
        self,
        shape: int | Iterable[int, ...] = (),
        name: str | None = None,
        var_id: int | None = None,
        **kwargs: dict,
    ) -> None:
        self._saddle_expr = None
        super().__init__(shape, name, var_id, **kwargs)

    @property
    def expr(self) -> dsp.saddle_min | dsp.saddle_max:
        assert self._saddle_expr is not None, "Must be associated with an SE."
        assert isinstance(self._saddle_expr, (dsp.saddle_min, dsp.saddle_max))
        return self._saddle_expr

    @expr.setter
    def expr(self, new_value: dsp.saddle_min | dsp.saddle_max) -> None:
        assert isinstance(new_value, (dsp.saddle_min, dsp.saddle_max))
        if self._saddle_expr is not None:
            if self._saddle_expr is not new_value:
                raise LocalVariableError("Cannot assign a local variable to multiple SEs.")
            else:
                return
        self._saddle_expr = new_value

    @property
    def value(self) -> np.ndarray:
        expr = self.expr
        if self._value is None:
            expr.numeric(
                None
            )  # TODO: allow numeric to take an arbitrary value rather than requiring a solve.
        return self._value  # can be none if numeric had no variable value

    @value.setter
    def value(self, new_value: np.ndarray) -> None:
        self._value = new_value

    def __repr__(self) -> str:
        return f"Local{super().__repr__()}"



import cvxpy as cp
import numpy as np

from dsp.problem import MinimizeMaximize, SaddlePointProblem
from dsp.saddle_atoms import inner, quasidef_quad_form


def test_quasidef_quad_form():
    x = cp.Variable(2, name="x")
    y = cp.Variable(2, name="y")

    P = np.eye(2)
    Q = -np.eye(2)
    S = np.eye(2)

    f_explicit = cp.quad_form(x, P) - cp.quad_form(y, -Q) + 2 * inner(x, S @ y)

    constraints = [cp.sum(x) <= 1, cp.sum(y) <= 1, x >= 0, y >= 0]

    saddle_problem = SaddlePointProblem(MinimizeMaximize(f_explicit), constraints)
    saddle_problem.solve()

    ref_value = saddle_problem.value
    ref_x = x.value
    ref_y = y.value

    f = quasidef_quad_form(x, y, P, Q, S)
    saddle_problem = SaddlePointProblem(MinimizeMaximize(f), constraints)
    saddle_problem.solve()

    assert np.isclose(saddle_problem.value, ref_value, atol=1e-4)
    assert np.allclose(x.value, ref_x, atol=1e-4)
    assert np.allclose(y.value, ref_y, atol=1e-4)  # unclear why these arent exactly equal
    assert np.isclose(f.value, ref_value, atol=1e-4)
    assert f.name().startswith("quasidef_quad_form(x, y")

    x.value = ref_x
    y.value = ref_y

    min_val = cp.Problem(cp.Minimize(f.get_convex_expression()), constraints).solve()
    max_val = cp.Problem(cp.Maximize(f.get_concave_expression()), constraints).solve()

    assert np.isclose(min_val, ref_value, atol=1e-4)
    assert np.isclose(max_val, ref_value, atol=1e-4)



import cvxpy as cp
import pytest

import dsp
from dsp.problem import MinimizeMaximize


def test_float_obj():
    obj = MinimizeMaximize(0)
    assert isinstance(obj.expr, cp.Expression)


def test_value():
    obj = MinimizeMaximize(cp.Constant(5))
    assert obj.value == 5


def test_invalid_argument():
    with pytest.raises(AssertionError):
        MinimizeMaximize("hello")


def test_properties():
    x = cp.Variable()
    y = cp.Variable()
    f = dsp.saddle_inner(x, y)

    obj = MinimizeMaximize(f)
    assert obj.is_dsp()
    assert obj.value is None
    assert obj.variables() == [x, y]
    assert obj.parameters() == []
    assert obj.atoms() == [dsp.saddle_inner]

    x.value = 1
    y.value = 2
    assert obj.value == 2



import cvxpy as cp
import numpy as np
import pytest

import dsp
from dsp import MinimizeMaximize, SaddlePointProblem, weighted_log_sum_exp
from dsp.parser import DSPError, Parser


def test_sum():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y")
    parser = Parser({x}, {y})
    expr = x + y
    parser.parse_expr_variables(expr, switched=False)

    assert parser.convex_vars == {x}
    assert parser.concave_vars == {y}


def test_negative_sum():
    x = cp.Variable()
    y = cp.Variable()
    parser = Parser({x}, {y})
    expr = -(x + y)
    parser.parse_expr_variables(expr, switched=False)
    assert parser.convex_vars == {x}
    assert parser.concave_vars == {y}


def test_mul_and_add():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    parser = Parser({x}, {y})
    wlse = weighted_log_sum_exp(x, y)

    expr = 1 + (-2) * (2 * (-wlse + 1))
    parser.parse_expr_variables(expr, switched=False)
    assert parser.convex_vars == {x}
    assert parser.concave_vars == {y}


def test_vars():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    z = cp.Variable(name="z")
    obj = dsp.MinimizeMaximize(weighted_log_sum_exp(x, y) + cp.exp(z))
    prob = SaddlePointProblem(obj, [y == 1, x == 1, z == 1])
    assert set(prob.convex_variables()) == {x, z}
    assert set(prob.concave_variables()) == {y}


def test_curvature_lumping():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    obj_ccv = MinimizeMaximize(cp.log(y) + x)
    prob = SaddlePointProblem(obj_ccv, [y == 1, x == 1])
    with pytest.raises(DSPError, match="specify"):
        prob.solve()

    prob = SaddlePointProblem(obj_ccv, [y == 1, x == 1], minimization_vars=[x])
    prob.solve()
    assert np.isclose(prob.x_prob.value, 1)


def test_just_affine():
    a = np.array([1, 2, 3])
    x = cp.Variable(3, name="x")
    y = cp.Variable(name="y")

    f = a @ x + cp.log(y) + cp.exp(x[0])
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [x >= 0, y >= 0])
    assert prob.is_dsp()

    f = np.ones(4) @ cp.hstack((x, y)) + cp.log(y) + cp.exp(x[0])
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [x >= 0, y >= 0])
    assert prob.is_dsp()  # Appears we can lump the variables despite different curvatures.

    f = cp.log(y) + cp.exp(x[0]) + np.ones(4) @ cp.hstack((x, y))
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [x >= 0, y >= 0])
    assert prob.is_dsp()  # Works independent of order.

    f = np.ones(4) @ cp.hstack((x, y))
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [x >= 0, y >= 0])
    assert not prob.is_dsp()  # Should fail if no implied curvatures.


@pytest.mark.parametrize("divisor", [-2, 2])
def test_divide(divisor):
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    f = weighted_log_sum_exp(x, y) / divisor
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [y == 1, x == 1])
    assert prob.is_dsp()

    prob.solve()
    assert np.isclose(prob.value, 1 / divisor)


def test_overlapping_vars():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    f = x + y
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(
        obj, [y == 1, x == 1], minimization_vars=[x], maximization_vars=[x, y]
    )
    assert not prob.is_dsp()


def test_unknown_curvature():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    f = cp.square(cp.log(x))
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [y == 1, x == 1], minimization_vars=[x])
    assert not prob.is_dsp()


def test_parser_no_constraints():
    x = cp.Variable(name="x")
    y = cp.Variable(name="y", nonneg=True)
    f = 0
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj)
    parser = Parser({x}, {y})

    assert not parser.x_constraints
    assert not parser.y_constraints


def test_no_mul_affine():
    x = cp.Variable(2, name="x")
    y = cp.Variable(2, name="y")
    f = x @ y
    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [y == 1, x == 1])
    assert not prob.is_dsp()



import cvxpy as cp
import numpy as np
import pytest

from dsp import saddle_quad_form
from dsp.problem import MinimizeMaximize, SaddlePointProblem


@pytest.mark.parametrize(
    "x_val,P_val", [([1, 0], [[1, 0], [0, 1]]), ([1, 1], [[1, 0.2], [0.2, 1]])]
)
def test_quad_form_equality(x_val, P_val):
    x = cp.Variable(2, name="x")
    P = cp.Variable((2, 2), name="P", PSD=True)

    x_val = np.array(x_val)
    P_val = np.array(P_val)

    f = saddle_quad_form(x, P)

    obj = MinimizeMaximize(f)
    prob = SaddlePointProblem(obj, [P == P_val, x == x_val])
    prob.solve()

    assert np.isclose(prob.value, x_val.T @ P_val @ x_val)


def test_quad_form_inequality():
    n = 2
    x = cp.Variable(n, name="x")
    P = cp.Variable((n, n), name="P", PSD=True)

    f = saddle_quad_form(x, P)
    assert f.value is None

    obj = MinimizeMaximize(f)
    P_val = np.eye(n)
    constraints = [P << P_val, cp.sum(x) == 1, x >= 0]
    prob = SaddlePointProblem(obj, constraints)
    prob.solve()

    p_val = P.value
    x_val = x.value

    # validate
    P_obj = x_val.T @ P @ x_val
    P_prob = cp.Problem(cp.Maximize(P_obj), [P << P_val])
    P_prob.solve()

    x_obj = cp.quad_form(x, p_val)
    x_prob = cp.Problem(cp.Minimize(x_obj), [cp.sum(x) == 1, x >= 0])
    x_prob.solve()

    assert np.isclose(P_prob.value, x_prob.value)

    x.value = x_val
    P.value = P_val

    min_val = cp.Problem(cp.Minimize(f.get_convex_expression()), constraints).solve()
    max_val = cp.Problem(cp.Maximize(f.get_concave_expression()), constraints).solve()

    assert np.isclose(min_val, x_prob.value)
    assert np.isclose(max_val, x_prob.value)


def test_value():
    n = 2
    x = cp.Variable(n)
    P = cp.Variable((n, n), PSD=True)

    saddle_quad = saddle_quad_form(x, P)

    assert saddle_quad.value is None
    assert saddle_quad.is_nonneg()
    assert not saddle_quad.is_incr(0)
    assert not saddle_quad.is_incr(1)

    x.value = np.arange(n)
    P.value = np.eye(n)

    assert saddle_quad.value == x.value.T @ P.value @ x.value


def test_saddle_quad_form_affine_constraint():

    n = 2
    sigma1 = np.eye(n)

    x = cp.Variable(n)
    P = cp.Variable((n, n), PSD=True)
    Delta = cp.Variable((n, n), name="Delta")

    f = saddle_quad_form(x, P)

    constraints = [P == sigma1 + Delta, Delta == 0, x == 1]

    problem = SaddlePointProblem(MinimizeMaximize(f), constraints)
    problem.solve()



from __future__ import annotations

import numpy as np


def np_vec(x: float | np.ndarray, order: str = "F") -> np.ndarray:
    """
    Convert a **scalar** or ndarray to a 1D array.
    """
    return np.atleast_1d(x).flatten(order=order)



from __future__ import annotations

import itertools
from abc import abstractmethod
from typing import Iterable

import cvxpy as cp
import numpy as np
from cvxpy.atoms.atom import Atom

import dsp
from dsp.cone_transforms import LocalToGlob
from dsp.local import LocalVariableError
from dsp.parser import DSPError, Parser, initialize_parser


class SaddleExtremum(Atom):
    def __init__(
        self,
        f: cp.Expression,
        constraints: Iterable[cp.Constraint],
    ) -> None:

        self.f = f
        self._parser = None

        self._validate_arguments(constraints)
        self.constraints = list(constraints)
        self.other_variables = [v for v in f.variables() if not isinstance(v, dsp.LocalVariable)]

        super().__init__(*self.other_variables)

    def _validate_arguments(self, constraints: list[cp.Constraint]) -> None:
        assert self.f.size == 1
        assert isinstance(constraints, Iterable)
        for c in constraints:
            for v in c.variables():
                if not isinstance(v, dsp.LocalVariable):
                    raise LocalVariableError(
                        "All variables in constraints must be instances of LocalVariable."
                    )
        assert isinstance(self.f, cp.Expression)

    def is_dsp(self) -> bool:
        try:
            self.parser  # noqa
            return all([c.is_dcp() for c in self.constraints])
        except DSPError:
            return False

    def is_incr(self, idx: int) -> bool:
        return False

    def is_decr(self, idx: int) -> bool:
        return False

    def shape_from_args(self) -> tuple[int, ...]:
        """Returns the (row, col) shape of the expression."""
        return self.f.shape

    def sign_from_args(self) -> tuple[bool, bool]:
        is_positive = self.f.is_nonneg()
        is_negative = self.f.is_nonpos()

        return is_positive, is_negative

    @abstractmethod
    def convex_variables(self) -> list[cp.Variable]:
        raise NotImplementedError

    @abstractmethod
    def concave_variables(self) -> list[cp.Variable]:
        raise NotImplementedError


class saddle_max(SaddleExtremum):
    r"""sup_{y\in Y}f(x,y)"""

    def __init__(
        self,
        f: cp.Expression,
        constraints: Iterable[cp.Constraint],
    ) -> None:

        super().__init__(f, constraints)

        self._concave_vars = set(filter(lambda v: isinstance(v, dsp.LocalVariable), f.variables()))
        self._concave_vars |= set(itertools.chain.from_iterable(c.variables() for c in constraints))

        for v in self._concave_vars:
            v.expr = self

    def convex_variables(self) -> list[cp.Variable]:
        return sorted(self.parser.convex_vars, key=lambda x: x.id)

    def concave_variables(self) -> list[cp.Variable]:
        return sorted(self._concave_vars, key=lambda x: x.id)

    @property
    def parser(self) -> Parser:
        if self._parser is None:

            parser = initialize_parser(
                self.f,
                minimization_vars=self.other_variables,
                maximization_vars=self._concave_vars,
                constraints=self.constraints,
            )

            assert set(self.other_variables) == parser.convex_vars
            assert set(self._concave_vars) == parser.concave_vars

            for v in self._concave_vars:
                v.expr = self

            self._parser = parser
            return parser
        else:
            return self._parser

    def name(self) -> str:
        return f"saddle_max({self.f.name()}, [{''.join([str(c) for c in self.constraints])}])"

    def numeric(self, values: list) -> np.ndarray | None:
        r"""
        Compute sup_{y\in Y}f(x,y) numerically
        """

        local_to_glob_y = LocalToGlob(self.convex_variables(), self.concave_variables())

        K_repr = self.parser.parse_expr_repr(self.f, switched=False, local_to_glob=local_to_glob_y)

        ccv = K_repr.concave_expr(values)
        if ccv is None:
            return None
        else:
            aux_problem = cp.Problem(cp.Maximize(ccv), self.constraints)
            aux_problem.solve()
            return aux_problem.value

    def is_atom_convex(self) -> bool:
        """Is the atom convex?"""
        return self.is_dsp()

    def is_atom_concave(self) -> bool:
        """Is the atom concave?"""
        return False


class saddle_min(SaddleExtremum):
    r"""inf_{x\in X}f(x,y)"""

    def __init__(
        self,
        f: cp.Expression,
        constraints: Iterable[cp.Constraint],
    ) -> None:

        super().__init__(f, constraints)

        self._convex_vars = set(filter(lambda v: isinstance(v, dsp.LocalVariable), f.variables()))
        self._convex_vars |= set(itertools.chain.from_iterable(c.variables() for c in constraints))

        for v in self._convex_vars:
            v.expr = self

    def concave_variables(self) -> list[cp.Variable]:
        return sorted(self.parser.concave_vars, key=lambda x: x.id)

    def convex_variables(self) -> list[cp.Variable]:
        return sorted(self._convex_vars, key=lambda x: x.id)

    @property
    def parser(self) -> Parser:
        if self._parser is None:
            parser = initialize_parser(
                self.f,
                minimization_vars=self._convex_vars,
                maximization_vars=self.other_variables,
                constraints=self.constraints,
            )

            assert set(self._convex_vars) == parser.convex_vars
            assert set(self.other_variables) == parser.concave_vars

            for v in self._convex_vars:
                v.expr = self

            self._parser = parser
            return parser
        else:
            return self._parser

    def name(self) -> str:
        return f"saddle_min({self.f.name()}, [{''.join([str(c) for c in self.constraints])}])"

    def numeric(self, values: list) -> np.ndarray | None:
        r"""
        Compute inf_{x\in X}f(x,y) numerically via -sup_{x\in X}-f(x,y)
        """
        neg_parser = initialize_parser(
            -self.f,
            minimization_vars=self.concave_variables(),
            maximization_vars=self.convex_variables(),
            constraints=self.constraints,
        )
        neg_local_to_glob = LocalToGlob(neg_parser.convex_vars, neg_parser.concave_vars)
        neg_K_repr = neg_parser.parse_expr_repr(
            -self.f, switched=False, local_to_glob=neg_local_to_glob
        )

        ccv = neg_K_repr.concave_expr(values)
        if ccv is None:
            return None
        else:
            aux_problem = cp.Problem(cp.Minimize(-ccv), self.constraints)
            aux_problem.solve()
            return aux_problem.value

    def is_atom_convex(self) -> bool:
        """Is the atom convex?"""
        return False

    def is_atom_concave(self) -> bool:
        """Is the atom concave?"""
        return self.is_dsp()


class conjugate(saddle_max):
    r"""f^*(y) = sup_{x\in \dom f} (y^Tx - f(x))"""

    def __init__(self, f: cp.Expression) -> None:
        assert isinstance(f, cp.Expression)

        self.original_f = f
        x_vars = f.variables()

        if not all(isinstance(x, dsp.LocalVariable) for x in x_vars):
            raise LocalVariableError("All conjugate variables must be local variables.")

        y_vars = [cp.Variable(name=f"{x.name()}_conjugate", shape=x.shape) for x in x_vars]

        obj = -f
        for x, y in zip(x_vars, y_vars):
            obj += dsp.inner(cp.vec(y), cp.vec(x))

        super().__init__(obj, [])

    def name(self) -> str:
        return f"conjugate({self.original_f.name()})"



import pytest

from dsp import *  # notational convenience
from dsp.local import LocalVariableError


def test_creating_saddle_function_and_solve_matrix_game():

    # ## Begin snippet 1

    import cvxpy as cp
    import numpy as np

    # from dsp import * # notational convenience

    x = cp.Variable(2)
    y = cp.Variable(2)
    C = np.array([[1, 2], [3, 1]])

    f = inner(x, C @ y)

    f.is_dsp()  # True

    f.convex_variables()  # [x]
    f.concave_variables()  # [y]
    f.affine_variables()  # []

    # ## End snippet 1

    assert f.is_dsp()
    assert f.convex_variables() == [x]
    assert f.concave_variables() == [y]
    assert f.affine_variables() == []

    # ## Begin snippet 2

    obj = MinimizeMaximize(f)
    constraints = [x >= 0, cp.sum(x) == 1, y >= 0, cp.sum(y) == 1]
    prob = SaddlePointProblem(obj, constraints)

    prob.is_dsp()  # True
    prob.convex_variables()  # [x]
    prob.concave_variables()  # [y]
    prob.affine_variables()  # []

    prob.solve()  # solves the problem
    prob.value  # 1.6666666666666667
    x.value  # array([0.66666667, 0.33333333])
    y.value  # array([0.33333333, 0.66666667])

    # ## End snippet 2

    assert prob.is_dsp()
    assert prob.convex_variables() == [x]
    assert prob.concave_variables() == [y]
    assert prob.affine_variables() == []
    assert np.isclose(prob.value, 1.6666666666666667)
    assert np.allclose(x.value, [0.66666667, 0.33333333])
    assert np.allclose(y.value, [0.33333333, 0.66666667])


def test_creating_a_saddle_max_and_solve_problem():

    import cvxpy as cp
    import numpy as np

    C = np.array([[1, 2], [3, 1]])

    # ## Begin snippet 3

    # Creating variables
    x = cp.Variable(2)

    # Creating local variables
    y_loc = LocalVariable(2)

    # Convex in x, concave in y_loc
    f = saddle_inner(x, C @ y_loc)

    # maximizes over y_loc
    G = saddle_max(f, [y_loc >= 0, cp.sum(y_loc) == 1])

    # ## End snippet 3

    # ## Begin snippet 4

    prob = cp.Problem(cp.Minimize(G), [x >= 0, cp.sum(x) == 1])

    prob.is_dsp()  # True

    prob.solve()  # solving the problem
    prob.value  # 1.6666666666666667
    x.value  # array([0.66666667, 0.33333333])

    # ## End snippet 4

    assert prob.is_dsp()
    assert np.isclose(prob.value, 1.6666666666666667)
    assert np.allclose(x.value, [0.66666667, 0.33333333])


def test_is_dsp_f1():
    import cvxpy as cp

    x = cp.Variable(name="x")
    z = cp.Variable(name="z")
    y_loc = LocalVariable(name="y_loc")

    f_1 = saddle_max(inner(x, y_loc) + z, [y_loc <= 1])
    assert f_1.is_dsp()

    assert set(f_1.convex_variables()) == {x, z}
    assert set(f_1.concave_variables()) == {y_loc}


def test_is_dsp_f2():
    import cvxpy as cp

    x = cp.Variable(name="x")
    y_loc = LocalVariable(name="y_loc")
    z_loc = LocalVariable(name="z_loc")

    f_2 = saddle_max(inner(x, y_loc) + z_loc, [y_loc <= 1, z_loc <= 1])
    assert f_2.is_dsp()

    assert set(f_2.convex_variables()) == {x}
    assert set(f_2.concave_variables()) == {y_loc, z_loc}


def test_is_dsp_f3():
    import cvxpy as cp

    x = cp.Variable(name="x")
    y_loc = LocalVariable(name="y_loc")
    z = cp.Variable(name="z")

    with pytest.raises(LocalVariableError):
        saddle_max(inner(x, y_loc) + z, [y_loc <= 1, z <= 1])


def test_is_dsp_f4():
    import cvxpy as cp

    x = cp.Variable(name="x")
    y = cp.Variable(name="y")
    y_loc = LocalVariable(name="y_loc")
    z_loc = LocalVariable(name="z_loc")

    f_4 = saddle_max(inner(x, y) + z_loc, [y_loc <= 1, z_loc <= 1])
    assert not f_4.is_dsp()
